#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
from dispatches.case_studies.renewables_case.wind_battery_PEM_LMP import (
    wind_battery_pem_mp_block,
    wind_battery_pem_variable_pairs,
    wind_battery_pem_periodic_variable_pairs,
    h2_mols_per_kg, pem_bar, pem_temp
)
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel
import pyomo.environ as pyo
import pandas as pd
from collections import deque
from functools import partial


def create_multiperiod_wind_pem_model(n_time_points, wind_cfs, input_params):
    """This function creates a MultiPeriodModel for the wind PEM model.

    Args:
        n_time_points (int): number of time period for the model.
        wind_cfs (sequence of floats): capacity factors of wind plant
        input_params (dict): system and cost parameters required for the Wind + PEM flowsheet

    Returns:
        MultiPeriodModel: a MultiPeriodModel for the wind PEM model.
    """
    if input_params['batt_mw'] != 0:
        raise ValueError("This model is not used for hybrids with battery. Battery MW must be 0.")

    # create the multiperiod model object
    mp_wind_pem = MultiPeriodModel(
        n_time_points=n_time_points,
        process_model_func=partial(wind_battery_pem_mp_block, input_params=input_params, verbose=False),
        linking_variable_func=wind_battery_pem_variable_pairs,
        periodic_variable_func=wind_battery_pem_periodic_variable_pairs,
    )

    wind_capacity_factors = {t:
                            {'wind_resource_config': {
                                'capacity_factor': 
                                    [wind_cfs[t]]}} for t in range(n_time_points)}

    mp_wind_pem.build_multi_period_model(wind_capacity_factors)

    return mp_wind_pem


def transform_design_model_to_operation_model(
    mp_wind_pem,
    wind_capacity=200e3,
    pem_power_capacity=25e3,
):
    """Transform the multiperiod wind PEM design model to operation model.

    Fix the sizes and deactivate periodic boundary condition

    Args:
        mp_wind_pem (MultiPeriodModel): a created multiperiod wind PEM object
        wind_capacity (float, optional): wind farm capapcity in KW. Defaults to 200e3.
        PEM (float, optional): PEM power output capacity in KW. Defaults to 25e3.
    """

    m = mp_wind_pem.pyomo_model
    blks = mp_wind_pem.get_active_process_blocks()
    m.pem_system_capacity = pyo.Var(domain=pyo.NonNegativeReals, initialize=pem_power_capacity, units=pyo.units.kW)

    for t, b in enumerate(blks):
        b.fs.windpower.system_capacity.fix(wind_capacity)
        b.fs.pem.max_p = pyo.Constraint(expr=b.fs.pem.electricity[0] <= m.pem_system_capacity)
        # deactivate periodic boundary condition
        if t == len(blks) - 1:
            b.periodic_constraints[0].deactivate()

    return


def update_wind_capacity_factor(mp_wind_pem, new_capacity_factors):
    """Update the wind capacity factor in the model during rolling horizon.

    Args:
        mp_wind_pem (MultiPeriodModel): a created multiperiod wind PEM object
        new_capacity_factors (list): a list of new wind capacity
    """

    blks = mp_wind_pem.get_active_process_blocks()
    for idx, b in enumerate(blks):
        b.fs.windpower.capacity_factor[0] = new_capacity_factors[idx]

    return


class MultiPeriodWindPEM:
    def __init__(
        self,
        model_data,
        wind_capacity_factors,
        wind_pmax_mw=200.0,
        pem_pmax_mw=25.0,
    ):
        """Initialize a multiperiod wind PEM model object for double loop.

        Args:
            model_data (GeneratorModelData): a GeneratorModelData that holds the generators params.
            wind_capacity_factors (list): a list of wind capacity.
            wind_pmax_mw (float, optional): wind farm capapcity in MW. Defaults to 200.
            pem_pmax_mw (float, optional): PEM power output capapcity in MW. Defaults to 25.

        Raises:
            ValueError: if wind capacity factor is not provided, ValueError will be raised
        """

        self.model_data = model_data

        if wind_capacity_factors is None:
            raise ValueError("Please provide wind capacity factors.")
        self._wind_capacity_factors = wind_capacity_factors

        self._wind_pmax_mw = wind_pmax_mw
        self._pem_pmax_mw = pem_pmax_mw

        # a list that holds all the result in pd DataFrame
        self.result_list = []

    def populate_model(self, b, horizon):
        """Create a wind-PEM model using the `MultiPeriod` package.

        The flowsheet model has option for battery; makes sure the battery size is 0 for Wind + PEM

        Args:
            b (block): this is an empty block passed in from eithe a bidder or tracker
            horizon (int): the number of time periods
        """
        blk = b
        if not blk.is_constructed():
            blk.construct()

        input_params = {
            'wind_mw': self._wind_pmax_mw,
            'pem_mw': self._pem_pmax_mw,
            "pem_bar": pem_bar,
            "pem_temp": pem_temp,
            'batt_mw': 0
        }
        blk.windPEM = create_multiperiod_wind_pem_model(horizon, wind_cfs=self._wind_capacity_factors[0:horizon], input_params=input_params)
        transform_design_model_to_operation_model(
            mp_wind_pem=blk.windPEM,
            wind_capacity=self._wind_pmax_mw * 1e3,
            pem_power_capacity=self._pem_pmax_mw * 1e3,
        )

        # deactivate any objective functions
        for obj in blk.windPEM.component_objects(pyo.Objective):
            obj.deactivate()

        # initialize time index for this block
        b._time_idx = pyo.Param(initialize=0, mutable=True)

        active_blks = blk.windPEM.get_active_process_blocks()

        # create expression that references underlying power variables in multi-period rankine
        blk.HOUR = pyo.Set(initialize=range(horizon))
        blk.P_T = pyo.Expression(blk.HOUR)
        blk.tot_cost = pyo.Expression(blk.HOUR)
        blk.wind_waste = pyo.Expression(blk.HOUR)

        for (t, b) in enumerate(active_blks):
            blk.P_T[t] = b.fs.splitter.grid_elec[0] * 1e-3
            blk.wind_waste[t] = b.fs.windpower.system_capacity * b.fs.windpower.capacity_factor[0] - b.fs.windpower.electricity[0]
            blk.tot_cost[t] = (b.fs.windpower.system_capacity * b.fs.windpower.op_cost / 8760 
                             + blk.windPEM.pem_system_capacity * b.fs.pem.op_cost / 8760 + b.fs.pem.var_cost
                             + blk.wind_waste[t])
        return

    def update_model(self, b, realized_h2_sales):
        """Update variables using future wind capacity and the realized hydrogen sales.

        The hydrogen sales aren't stored at the moment.

        Args:
            b (block): the block that needs to be updated
        """

        blk = b
        mp_wind_pem = blk.windPEM

        # shift the time -> update capacity_factor
        time_advance = min(len(realized_h2_sales), 24)
        b._time_idx = pyo.value(b._time_idx) + time_advance

        new_capacity_factors = self._get_capacity_factors(b)
        update_wind_capacity_factor(mp_wind_pem, new_capacity_factors)

        return

    def _get_capacity_factors(self, b):
        """Fetch the future capacity factor.

        Args:
            b (block): the block that needs to be updated

        Returns:
            list: the capcity factors for the immediate future
        """

        horizon_len = len(b.windPEM.get_active_process_blocks())
        ans = self._wind_capacity_factors[
            pyo.value(b._time_idx) : pyo.value(b._time_idx) + horizon_len
        ]

        return ans

    @staticmethod
    def get_last_delivered_power(b, last_implemented_time_step):
        """Get last delivered power.

        Args:
            b (block): a multiperiod block
            last_implemented_time_step (int):  time index for the last implemented time period

        Returns:
            float: last delivered power
        """

        blk = b
        return pyo.value(blk.P_T[last_implemented_time_step])

    @staticmethod
    def get_implemented_profile(b, last_implemented_time_step):
        """Get implemented profiles, i.e., realized hydrogen sales.

        Args:
            b (block): a multiperiod block
            last_implemented_time_step (int):  time index for the last implemented time period

        Returns:
            dict: dictionalry of implemented profiles.
        """

        blk = b
        mp_wind_pem = blk.windPEM
        active_blks = mp_wind_pem.get_active_process_blocks()

        realized_h2_sales = deque(
            pyo.value(active_blks[t].fs.pem.outlet.flow_mol[0] / h2_mols_per_kg * 3600)
            for t in range(last_implemented_time_step + 1)
        )

        return {
            "realized_h2_sales": realized_h2_sales,
        }

    def record_results(self, b, date=None, hour=None, **kwargs):
        """Record the operations stats for the model, i.e., generator name, data, hour, horizon,
        total wind generation, total power output, wind power output, wind to PEM output, hydrogen sales, total costs.

        Args:
            b (block): a multiperiod block
            date (str, optional): current simulation date. Defaults to None.
            hour (int, optional): current simulation hour. Defaults to None.
        """
        blk = b
        mp_wind_pem = blk.windPEM
        active_blks = mp_wind_pem.get_active_process_blocks()

        df_list = []
        for t, process_blk in enumerate(active_blks):

            result_dict = {}

            result_dict["Generator"] = self.model_data.gen_name
            result_dict["Date"] = date
            result_dict["Hour"] = hour

            # simulation inputs
            result_dict["Horizon [hr]"] = int(t)

            # model vars
            result_dict["Total Wind Generation [MW]"] = float(
                round(pyo.value(process_blk.fs.windpower.electricity[0]) * 1e-3, 2)
            )
            result_dict["Total Power Output [MW]"] = float(
                round(pyo.value(blk.P_T[t]), 2)
            )
            result_dict["Wind Power Output [MW]"] = float(
                round(pyo.value(process_blk.fs.splitter.grid_elec[0] * 1e-3), 2)
            )
            result_dict["Wind to PEM [MW]"] = float(
                round(pyo.value(process_blk.fs.splitter.pem_elec[0] * 1e-3), 2)
            )
            result_dict["Wind Curtailment [MW]"] = float(
                round(pyo.value(blk.wind_waste[0]), 2)
            )
            result_dict["Hydrogen Sales [kg]"] = float(
                round(pyo.value(process_blk.fs.pem.outlet.flow_mol[0] / h2_mols_per_kg * 3600), 2)
            )
            result_dict["Total Cost [$]"] = float(round(pyo.value(blk.tot_cost[t]), 2))

            for key in kwargs:
                result_dict[key] = kwargs[key]

            result_df = pd.DataFrame.from_dict(result_dict, orient="index")
            df_list.append(result_df.T)

        # append to result list
        self.result_list.append(pd.concat(df_list))

        return

    def write_results(self, path):
        """Write the saved results to a csv file.

        Args:
            path (str): the path to write the results.
        """

        pd.concat(self.result_list).to_csv(path, index=False)

    @property
    def power_output(self):
        return "P_T"

    @property
    def total_cost(self):
        return ("tot_cost", 1)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################


# General python imports
import json
import omlt
import pickle

from pyomo.environ import (
    ConcreteModel, 
    Var, 
    value, 
    Block, 
    Expression, 
    Constraint,
    Objective, 
    NonNegativeReals,
    sqrt,
)

# Import IDAES  and DISPATCHES components
from idaes.core.util import to_json, from_json
from idaes.core.solvers import get_solver

import dispatches.case_studies.simple_rankine_cycle.simple_rankine_cycle as src


# load scaling and bounds for each surrogate
with open("training_parameters_revenue.json", 'rb') as f:
    rev_data = json.load(f)

with open("training_parameters_zones.json", 'rb') as f:
    zone_data = json.load(f)

with open("training_parameters_nstartups.json", 'rb') as f:
    nstartups_data = json.load(f)

# load scikit neural networks
with open("scikit_models/revenue.pkl", 'rb') as f:
    nn_revenue = pickle.load(f)

with open("scikit_models/zones.pkl", 'rb') as f:
    nn_zones = pickle.load(f)

with open("scikit_models/nstartups.pkl", 'rb') as f:
    nn_nstartups = pickle.load(f)

# Load scikit models and create OMLT NetworkDefinition objects
# Revenue model definition
input_bounds_rev = list(zip(rev_data['xmin'], rev_data['xmax']))
scaling_object_revenue = omlt.OffsetScaling(
    offset_inputs=rev_data['xm_inputs'],
    factor_inputs=rev_data['xstd_inputs'],
    offset_outputs=[rev_data['zm_revenue']],
    factor_outputs=[rev_data['zstd_revenue']],
)
net_rev_defn = load_scikit_mlp(nn_revenue, scaling_object_revenue, input_bounds_rev)

# Zone model definition
input_bounds_zones = list(zip(zone_data['xmin'], zone_data['xmax']))
scaling_object_zones = omlt.OffsetScaling(
    offset_inputs=zone_data['xm_inputs'],
    factor_inputs=zone_data['xstd_inputs'],
    offset_outputs=zone_data['zm_zones'],
    factor_outputs=zone_data['zstd_zones'],
)
net_zone_defn = load_scikit_mlp(nn_zones, scaling_object_zones, input_bounds_zones)

# Nstartup model definition
input_bounds_nstartups = list(zip(nstartups_data['xmin'], nstartups_data['xmax']))
scaling_object_nstartups = omlt.OffsetScaling(
    offset_inputs=nstartups_data['xm_inputs'],
    factor_inputs=nstartups_data['xstd_inputs'],
    offset_outputs=[nstartups_data['zm_nstartups']],
    factor_outputs=[nstartups_data['zstd_nstartups']],
)
net_nstartups_defn = load_scikit_mlp(nn_nstartups, scaling_object_nstartups, input_bounds_nstartups)


#Denote the scaled power output for each of the 10 zones (0 corresponds to pmin, 1.0 corresponds to pmax)
zone_outputs = [0.0, 0.15, 0.25, 0.35, 0.45, 0.55, 0.65, 0.75, 0.85, 1.0]

def conceptual_design_problem_nn(
    heat_recovery=False,
    calc_boiler_eff=False,
    p_lower_bound=10,
    p_upper_bound=500,
    capital_payment_years=5,
    plant_lifetime=20,
    coal_price=51.96,
):

    m = ConcreteModel()

    # Create capex plant
    m.cap_fs = src.create_model(
        heat_recovery=heat_recovery,
        capital_fs=True, 
        calc_boiler_eff=False,
    )
    src.set_inputs(m.cap_fs)
    src.initialize_model(m.cap_fs)
    src.close_flowsheet_loop(m.cap_fs)
    src.add_capital_cost(m.cap_fs)

    # capital cost (M$/yr)
    cap_expr = m.cap_fs.fs.capital_cost / capital_payment_years

    #surrogate market inputs (not technically part of rankine cycle model but are used in market model)
    m.pmin_multi = Var(within=NonNegativeReals, bounds=(0.15, 0.45), initialize=0.3)
    m.ramp_multi = Var(within=NonNegativeReals, bounds=(0.5, 1.0), initialize=0.75)
    m.min_up_time = Var(within=NonNegativeReals, bounds=(1.0, 16.0), initialize=4.0)
    m.min_dn_multi = Var(within=NonNegativeReals, bounds=(0.5, 2.0), initialize=1.0)
    m.marg_cst =  Var(within=NonNegativeReals, bounds=(5, 30), initialize=15)
    m.no_load_cst =  Var(within=NonNegativeReals, bounds=(0, 2.5), initialize=1)
    m.startup_cst = Var(within=NonNegativeReals, bounds=(0, 136), initialize=75)

    #actual generator values
    m.pmax = Expression(expr=m.cap_fs.fs.net_cycle_power_output * 1e-6)
    m.pmin = Expression(expr=m.pmin_multi * m.pmax)
    m.min_dn_time = Expression(expr=m.min_dn_multi * m.min_up_time)
    m.ramp_rate= Expression(expr=m.ramp_multi * (m.pmax - m.pmin))

    m.inputs = [
        m.pmax, m.pmin_multi, m.ramp_multi, m.min_up_time, m.min_dn_multi,
        m.marg_cst, m.no_load_cst, m.startup_cst,
    ]

    ######################################
    # Revenue surrogate
    ######################################
    m.rev_surrogate = Var()
    m.nn_rev = omlt.OmltBlock()

    # build the formulation on the omlt block
    formulation_rev = omlt.neuralnet.ReducedSpaceContinuousFormulation(net_rev_defn)
    m.nn_rev.build_formulation(
        formulation_rev, input_vars=m.inputs, output_vars=[m.rev_surrogate]
    )
    m.revenue = Expression(expr=0.5*sqrt(m.rev_surrogate**2 + 0.001**2) + 0.5*m.rev_surrogate)

    ######################################
    # nstartups surrogate
    ######################################
    m.nstartups_surrogate = Var()
    m.nn_nstartups = omlt.OmltBlock()

    #build the formulation on the omlt block
    formulation_nstartups = omlt.neuralnet.ReducedSpaceContinuousFormulation(net_nstartups_defn)
    m.nn_nstartups.build_formulation(
        formulation_nstartups, input_vars=m.inputs, output_vars=[m.nstartups_surrogate]
    )
    m.nstartups = Expression(expr=0.5*sqrt(m.nstartups_surrogate**2 + 0.001**2) + 0.5*m.nstartups_surrogate)

    ############################################
    #zone surrogates
    ############################################
    m.nn_zones = omlt.OmltBlock()
    formulation_zones = omlt.neuralnet.ReducedSpaceContinuousFormulation(net_zone_defn)
    m.zone_hours_surrogate = Var(range(11), within=NonNegativeReals)
    m.nn_zones.build_formulation(
        formulation_zones, input_vars=m.inputs, output_vars=list(m.zone_hours_surrogate.values())
    )

    # zone off flowsheet
    off_fs = Block()
    off_fs.fs = Block()
    off_fs.fs.operating_cost = m.no_load_cst * m.pmax
    off_fs.zone_hours = Expression(
        expr=0.5*sqrt(m.zone_hours_surrogate[0]**2 + 0.001**2) + 0.5*m.zone_hours_surrogate[0]
    )
    setattr(m, 'zone_{}'.format('off'), off_fs)

    #Create a surrogate flowsheet for each operating zone
    op_zones = []
    init_flag = 0
    for (i, zone_output) in enumerate(zone_outputs):
        print("Creating instance ", i)
        op_fs = src.create_model(
            heat_recovery=heat_recovery,
            capital_fs=False,
            calc_boiler_eff=calc_boiler_eff,
        )
        # Set model inputs for the capex and opex plant
        src.set_inputs(op_fs)

        # Fix the p_max of op_fs to p of cap_fs for initialization
        op_fs.fs.net_power_max.fix(value(m.cap_fs.fs.net_cycle_power_output))

        #initialize with json
        if init_flag == 0:
            # Initialize the opex plant
            src.initialize_model(op_fs)

            # save model state after initializing the first instance
            init_model = to_json(op_fs.fs, return_dict=True)
            init_flag = 1
        else:
            # Initialize the capex and opex plant
            from_json(op_fs.fs, sd=init_model)

        # Closing the loop in the flowsheet
        src.close_flowsheet_loop(op_fs)
        src.add_operating_cost(op_fs, coal_price=coal_price)

        # Unfix op_fs p_max and set constraint linking that to cap_fs p_max
        op_fs.fs.net_power_max.unfix()
        op_fs.fs.eq_p_max = Constraint(
            expr=op_fs.fs.net_power_max == m.cap_fs.fs.net_cycle_power_output * 1e-6
        )

        # Fix zone power output
        op_fs.fs.eq_fix_power = Constraint(
            expr=op_fs.fs.net_cycle_power_output * 1e-6 == zone_output * (m.pmax-m.pmin) + m.pmin
        )

        # Smooth max on zone hours (avoids negative hours)
        op_fs.zone_hours = Expression(
            expr=0.5*sqrt(m.zone_hours_surrogate[i+1]**2 + 0.001**2) + 0.5*m.zone_hours_surrogate[i+1]
        )

        #unfix the boiler flow rate
        op_fs.fs.boiler.inlet.flow_mol[0].setlb(0.01)
        op_fs.fs.boiler.inlet.flow_mol[0].unfix()
        setattr(m, 'zone_{}'.format(i), op_fs)
        op_zones.append(op_fs)

    # scale hours between 0 and 1 year (8736 hours were used in simulation)
    m.zone_total_hours = sum(
        op_zones[i].zone_hours for i in range(len(op_zones))
    ) + off_fs.zone_hours

    for op_fs in op_zones:
        op_fs.scaled_zone_hours = Var(within=NonNegativeReals, bounds=(0, 8736), initialize=100)
        # NOTE: scaled_hours_i = surrogate_i * 8736 / surrogate_total
        op_fs.con_scale_zone_hours = Constraint(
            expr=op_fs.scaled_zone_hours * m.zone_total_hours == op_fs.zone_hours * 8736
        )

    off_fs.scaled_zone_hours = Var(within=NonNegativeReals, bounds=(0, 8736), initialize=100)
    off_fs.con_scale_zone_hours = Constraint(
        expr=off_fs.scaled_zone_hours * m.zone_total_hours == off_fs.zone_hours * 8736
    )

    # Operating cost in $MM (million dollars)
    m.op_expr = sum(
        op_zones[i].scaled_zone_hours * op_zones[i].fs.operating_cost 
        for i in range(len(op_zones))
    ) * 1e-6 + off_fs.scaled_zone_hours * off_fs.fs.operating_cost * 1e-6

    # startup cost in MM$
    m.startup_expr = m.startup_cst * m.nstartups * m.pmax * 1e-6

    m.op_zones = op_zones

    #Piecewise cost limits, connect marginal cost to operating cost
    m.connect_mrg_cost = Constraint(
        expr=m.marg_cst == 0.5*(op_zones[0].fs.operating_cost/m.pmin + op_zones[-1].fs.operating_cost/m.pmax)
    )

    # Expression for total cap and op cost - $
    m.total_cost = Expression(
        expr=plant_lifetime * (m.op_expr + m.startup_expr) + capital_payment_years * cap_expr
    )

    # Expression for total revenue
    m.total_revenue = Expression(expr=plant_lifetime * m.revenue)

    # Objective $
    m.obj = Objective(expr=-(m.total_revenue - m.total_cost))

    # Unfixing the boiler inlet flowrate for capex plant
    m.cap_fs.fs.boiler.inlet.flow_mol[0].unfix()

    # Setting bounds for the capex plant flowrate
    m.cap_fs.fs.boiler.inlet.flow_mol[0].setlb(0.01)

    # Setting bounds for net cycle power output for the capex plant
    m.cap_fs.fs.eq_min_power = Constraint(
        expr=m.cap_fs.fs.net_cycle_power_output >= p_lower_bound * 1e6
    )

    m.cap_fs.fs.eq_max_power = Constraint(
        expr=m.cap_fs.fs.net_cycle_power_output <= p_upper_bound * 1e6)

    return m



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
# Import Pyomo libraries
from pyomo.environ import Var, Param, NonNegativeReals, units as pyunits
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData)
from idaes.core.util.initialization import solve_indexed_blocks
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("SolarPV", doc="Soar PV plant using capacity factors")
class SolarPVData(UnitModelBlockData):
    """
    Solar PV plant using resource data.
    Unit model to convert Solar PV resource into electricity.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Solar PV plant does not support dynamic models, thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag",
        doc="""Solar PV plant does not have defined volume, thus this must be False."""))
    CONFIG.declare("capacity_factor", ConfigValue(
        default=None,
        domain=ListOf(float),
        description="Capacity Factors per Timestep",
        doc="Capacity Factor in a list for each Timestep in model"))

    def build(self):
        """Building model

        Args:
            None
        Returns:
            None
        """
        super().build()

        self.system_capacity = Var(within=NonNegativeReals,
                                     initialize=0.0,
                                     doc="Rated system capacity of Solar PV farm",
                                     units=pyunits.kW)

        self.capacity_factor = Param(self.flowsheet().config.time,
                                     within=NonNegativeReals,
                                     mutable=True,
                                     initialize=0.0,
                                     doc="Ratio of power output to rated capacity, on annual or time series basis",
                                     units=pyunits.kW/pyunits.kW)

        self.electricity = Var(self.flowsheet().config.time,
                               within=NonNegativeReals,
                               initialize=0.0,
                               doc="Electricity production",
                               units=pyunits.kW)

        self.electricity_out = Port(noruleinit=True, doc="A port for electricity flow")
        self.electricity_out.add(self.electricity, "electricity")

        @self.Constraint(self.flowsheet().config.time)
        def elec_from_capacity_factor(b, t):
            return b.electricity[t] <= self.system_capacity * self.capacity_factor[t]

        self.setup_resource()

    def _get_performance_contents(self, time_point=0):
        return {"vars": {"Electricity": self.electricity[time_point].value}}

    def setup_resource(self):
        """
        """
        if not self.config.capacity_factor:
            raise ValueError("Config with `capacity_factor` must be provided using `default` argument")

        if not len(self.config.capacity_factor) >= len(self.flowsheet().config.time.data()):
            raise ValueError("Config with 'capacity_factor' must provide data per time step")
            
        for n, time in enumerate(self.flowsheet().config.time.data()):
            self.capacity_factor[time].set_value(self.config.capacity_factor[n])   

    def initialize_build(self, **kwargs):
        for t in self.flowsheet().config.time:
            self.electricity[t].set_value(self.system_capacity * self.capacity_factor[t])



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
import pytest
from sys import platform
pytest.importorskip("tensorflow", reason="optional dependencies for surrogate modeling not available")

from dispatches.case_studies.renewables_case.RE_surrogate_optimization_steadystate import *


def load_model():
    net_rev_defn, net_frequency_defn, dispatch_clusters_mean, pem_clusters_mean, resource_clusters_mean = load_surrogate_model(re_nn_dir)
    assert len([l for l in net_rev_defn.layers]) == 4
    assert len([l for l in net_frequency_defn.layers]) == 5
    assert dispatch_clusters_mean.mean() == pytest.approx(0.2445, rel=1e-3)
    assert pem_clusters_mean.mean() == pytest.approx(0.2215, rel=1e-3)
    assert resource_clusters_mean.mean() == pytest.approx(0.5547, rel=1e-3)


def test_RE_surrogate_steady_state_fixed():
    results = run_design(PEM_bid=30, PEM_size=200)
    assert results['e_revenue'] == pytest.approx(-5989492, rel=1e-3)
    assert results['h_revenue'] == pytest.approx(43656556, rel=1e-3)
    assert results['NPV'] == pytest.approx(-1353268767, rel=1e-3)


def test_RE_surrogate_steady_state():
    results = run_design()
    assert results['pem_mw'] == pytest.approx(350.02, rel=1e-3)
    assert results['pem_bid'] == pytest.approx(45.00, rel=1e-3)
    assert results['e_revenue'] == pytest.approx(5777449, rel=1e-3)
    assert results['h_revenue'] == pytest.approx(66398096, rel=1e-3)
    assert results['NPV'] == pytest.approx(-1186120046, rel=1e-3)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################


import os
import json

# Pyomo imports
from pyomo.environ import (
    ConcreteModel, 
    Var, 
    value, 
    Block, 
    Expression, 
    Constraint, 
    NonNegativeReals,
    sqrt,
    Objective,
)

# Use idaes SurrogateBlock
from idaes.core.util import to_json, from_json
from idaes.core.solvers import get_solver
from idaes.core.surrogate.alamopy import AlamoSurrogate
from idaes.core.surrogate.surrogate_block import SurrogateBlock

import dispatches.case_studies.simple_rankine_cycle.simple_rankine_cycle as src 

# load scaling and bounds for each surrogate
with open("training_parameters_revenue.json", 'rb') as f:
    rev_data = json.load(f)

with open("training_parameters_zones.json", 'rb') as f:
    zone_data = json.load(f)

with open("training_parameters_nstartups.json", 'rb') as f:
    nstartups_data = json.load(f)


xm = rev_data['xm_inputs']
xstd = rev_data['xstd_inputs']

zm_rev = rev_data['zm']
zstd_rev = rev_data['zstd']

zm_zones = zone_data['zm']
zstd_zones = zone_data['zstd']

zm_nstartups = nstartups_data['zm']
zstd_nstartups = nstartups_data['zstd']


# Load surrogates from alamo .json files
alamo_revenue = AlamoSurrogate.load_from_file("alamo_models/revenue.json")
alamo_nstartups = AlamoSurrogate.load_from_file("alamo_models/nstartups.json")
alamo_zones = AlamoSurrogate.load_from_file("alamo_models/zones.json")

# Denote the scaled power output for each of the 10 zones 
# (0 corresponds to pmin, 1.0 corresponds to pmax)
zone_outputs = [0.0, 0.15, 0.25, 0.35, 0.45, 0.55, 0.65, 0.75, 0.85, 1.0]

# this creates the alamo surrogate flowsheet of the rankine cycle
def conceptual_design_problem_alamo(
    heat_recovery=False,
    calc_boiler_eff=False,
    p_lower_bound=10,
    p_upper_bound=500,
    capital_payment_years=5,
    plant_lifetime=20,
    coal_price=51.96,
):

    m = ConcreteModel()

    # Create capex plant
    m.cap_fs = src.create_model(
        heat_recovery=heat_recovery,
        capital_fs=True, 
        calc_boiler_eff=False,
    )
    m.cap_fs = src.set_inputs(m.cap_fs)
    m.cap_fs = src.initialize_model(m.cap_fs)
    m.cap_fs = src.close_flowsheet_loop(m.cap_fs)
    m.cap_fs = src.add_capital_cost(m.cap_fs)

    # capital cost (M$/yr)
    cap_expr = m.cap_fs.fs.capital_cost/capital_payment_years

    # Surrogate market inputs (not technically part of rankine cycle 
    # model but are used in market model)
    m.pmin_multi = Var(within=NonNegativeReals, bounds=(0.15, 0.45), initialize=0.3)
    m.ramp_multi = Var(within=NonNegativeReals, bounds=(0.5, 1.0), initialize=0.5)
    m.min_up_time = Var(within=NonNegativeReals, bounds=(1.0, 16.0), initialize=4.0)
    m.min_dn_multi = Var(within=NonNegativeReals, bounds=(0.5, 2.0), initialize=1.0)
    m.marg_cst =  Var(within=NonNegativeReals, bounds=(5, 30), initialize=15)
    m.no_load_cst =  Var(within=NonNegativeReals, bounds=(0, 2.5), initialize=1)
    m.startup_cst = Var(within=NonNegativeReals, bounds=(0, 136), initialize=75)

    # actual generator values
    m.pmax = Expression(expr=m.cap_fs.fs.net_cycle_power_output * 1e-6)
    m.pmin = Expression(expr=m.pmin_multi * m.pmax)
    m.min_dn_time = Expression(expr=m.min_dn_multi * m.min_up_time)
    m.ramp_rate= Expression(expr=m.ramp_multi * (m.pmax - m.pmin))

    ######################################
    # revenue surrogate
    ######################################
    m.alamo_revenue_surrogate = SurrogateBlock()
    m.alamo_revenue_surrogate.build_model(alamo_revenue)
    input_dict = m.alamo_revenue_surrogate.input_vars_as_dict()

    m.pmax_revenue_in = Constraint(
        expr=input_dict["pmax"] == (m.pmax - xm[0]) / xstd[0]
    )
    m.pmin_revenue_in = Constraint(
        expr=input_dict["pmin_multi"] == (m.pmin_multi - xm[1]) / xstd[1]
    )
    m.ramp_revenue_in = Constraint(
        expr=input_dict["ramp_multi"] == (m.ramp_multi - xm[2]) / xstd[2]
    )
    m.min_up_revenue_in = Constraint(
        expr=input_dict["min_up_time"] == (m.min_up_time - xm[3]) / xstd[3]
    )
    m.min_dn_revenue_in = Constraint(
        expr=input_dict["min_down_multi"] == (m.min_dn_multi - xm[4]) / xstd[4]
    )
    m.marg_cst_revenue_in = Constraint(
        expr=input_dict["marg_cst"] == (m.marg_cst - xm[5]) / xstd[5]
    )
    m.no_load_cst_revenue_in = Constraint(
        expr=input_dict["no_load_cst"] == (m.no_load_cst - xm[6]) / xstd[6]
    )
    m.startup_cst_revenue_in = Constraint(
        expr=input_dict["startup_cst"] == (m.startup_cst - xm[7]) / xstd[7]
    )

    # Define a variable for revenue
    m.rev_surrogate = Var()
    m.revenue_out = Constraint(
        expr=m.rev_surrogate == m.alamo_revenue_surrogate.outputs['revenue'] * zstd_rev + zm_rev
    )

    # this is a smooth-max; it sets negative revenue to zero
    m.revenue = Expression(expr=0.5*sqrt(m.rev_surrogate**2 + 0.001**2) + 0.5*m.rev_surrogate)

    #######################################
    # nstartups surrogate
    #######################################
    m.alamo_nstartups_surrogate = SurrogateBlock()
    m.alamo_nstartups_surrogate.build_model(alamo_nstartups)
    input_dict = m.alamo_nstartups_surrogate.input_vars_as_dict()

    m.pmax_nstartups_in = Constraint(
        expr=input_dict["pmax"] == (m.pmax - xm[0]) / xstd[0]
    )
    m.pmin_nstartups_in = Constraint(
        expr=input_dict["pmin_multi"] == (m.pmin_multi - xm[1]) / xstd[1]
    )
    m.ramp_nstartups_in = Constraint(
        expr=input_dict["ramp_multi"] == (m.ramp_multi - xm[2]) / xstd[2]
    )
    m.min_up_nstartups_in = Constraint(
        expr=input_dict["min_up_time"] == (m.min_up_time - xm[3]) / xstd[3]
    )
    m.min_dn_nstartups_in = Constraint(
        expr=input_dict["min_down_multi"] == (m.min_dn_multi - xm[4]) / xstd[4]
    )
    m.marg_cst_nstartups_in = Constraint(
        expr=input_dict["marg_cst"] == (m.marg_cst - xm[5]) / xstd[5]
    )
    m.no_load_cst_nstartups_in = Constraint(
        expr=input_dict["no_load_cst"] == (m.no_load_cst - xm[6]) / xstd[6]
    )
    m.startup_cst_nstartups_in = Constraint(
        expr=input_dict["startup_cst"] == (m.startup_cst - xm[7]) / xstd[7]
    )

    # Define a variable for number of strartups
    m.nstartups_surrogate = Var()
    m.nstartups_out = Constraint(
        expr=m.nstartups_surrogate == m.alamo_nstartups_surrogate.outputs['nstartups']*zstd_nstartups + zm_nstartups
    )

    m.nstartups = Expression(expr=0.5*sqrt(m.nstartups_surrogate**2 + 0.001**2) + 0.5*m.nstartups_surrogate)

    ############################################
    # zone surrogates
    ############################################
    m.alamo_zones_surrogate = SurrogateBlock()
    m.alamo_zones_surrogate.build_model(alamo_zones)
    input_dict = m.alamo_zones_surrogate.input_vars_as_dict()
    output_dict = m.alamo_zones_surrogate.output_vars_as_dict()

    #the zone surrogate has 11 outputs: off + 10 zones from pmin to pmax
    m.zone_hours_surrogate = Var(range(11))

    m.pmax_zones_in = Constraint(expr=input_dict["pmax"] == (m.pmax - xm[0]) / xstd[0])
    m.pmin_zones_in = Constraint(expr=input_dict["pmin_multi"] == (m.pmin_multi - xm[1]) / xstd[1])
    m.ramp_zones_in = Constraint(expr=input_dict["ramp_multi"] == (m.ramp_multi - xm[2]) / xstd[2])
    m.min_up_zones_in = Constraint(expr=input_dict["min_up_time"] == (m.min_up_time - xm[3]) / xstd[3])
    m.min_dn_zones_in = Constraint(expr=input_dict["min_down_multi"] == (m.min_dn_multi - xm[4]) / xstd[4])
    m.marg_cst_zones_in = Constraint(expr=input_dict["marg_cst"] == (m.marg_cst - xm[5]) / xstd[5])
    m.no_load_cst_zones_in = Constraint(expr=input_dict["no_load_cst"] == (m.no_load_cst - xm[6]) / xstd[6])
    m.startup_cst_zones_in = Constraint(expr=input_dict["startup_cst"] == (m.startup_cst - xm[7]) / xstd[7])

    m.zone_hours_out = Constraint(range(11))
    for (i, k) in enumerate(output_dict.keys()):
        m.zone_hours_out[i] = m.zone_hours_surrogate[i] == m.alamo_zones_surrogate.outputs[k]*zstd_zones[i] + zm_zones[i]

    ############################################
    # begin flowsheets
    ############################################
    # zone off flowsheet
    off_fs = Block()
    off_fs.fs = Block()
    off_fs.fs.operating_cost = Expression(expr=m.no_load_cst * m.pmax)
    off_fs.zone_hours = Expression(expr=0.5*sqrt(m.zone_hours_surrogate[0]**2 + 0.001**2) + 0.5*m.zone_hours_surrogate[0])
    setattr(m, 'zone_{}'.format('off'), off_fs)

    # Create a surrogate flowsheet for each operating zone
    op_zones = []
    init_flag = 0
    for (i, zone_output) in enumerate(zone_outputs):
        print("Creating instance ", i)
        op_fs = src.create_model(
            heat_recovery=heat_recovery,
            capital_fs=False,
            calc_boiler_eff=calc_boiler_eff,
        )
        # Set model inputs for the capex and opex plant
        op_fs = src.set_inputs(op_fs)

        # Fix the p_max of op_fs to p of cap_fs for initialization
        op_fs.fs.net_power_max.fix(value(m.cap_fs.fs.net_cycle_power_output))

        # initialize with json. this speeds up model instantiation. it writes a json file \
        # for the first flowsheet which is used to initialize the next flowsheets
        if init_flag == 0:
            # Initialize the opex plant
            op_fs = src.initialize_model(op_fs)

            # save model state after initializing the first instance
            init_model = to_json(op_fs.fs, return_dict=True)
            init_flag = 1
        else:
            # Initialize the capex and opex plant
            from_json(op_fs.fs, sd=init_model)

        # Closing the loop in the flowsheet
        op_fs = src.close_flowsheet_loop(op_fs)
        op_fs = src.add_operating_cost(op_fs, coal_price=coal_price)

        # Unfix op_fs p_max and set constraint linking that to cap_fs p_max
        op_fs.fs.net_power_max.unfix()
        op_fs.fs.eq_p_max = Constraint(
            expr=op_fs.fs.net_power_max == m.cap_fs.fs.net_cycle_power_output*1e-6
        )

        # Fix zone power output
        op_fs.fs.eq_fix_power = Constraint(
            expr=op_fs.fs.net_cycle_power_output*1e-6 == zone_output*(m.pmax-m.pmin) + m.pmin
        )

        # smooth max on zone hours (avoids negative hours)
        op_fs.zone_hours = Expression(
            expr=0.5*sqrt(m.zone_hours_surrogate[i+1]**2 + 0.001**2) + 0.5*m.zone_hours_surrogate[i+1]
        )

        #unfix the boiler flow rate
        op_fs.fs.boiler.inlet.flow_mol[0].setlb(0.01)
        op_fs.fs.boiler.inlet.flow_mol[0].unfix()
        setattr(m, 'zone_{}'.format(i), op_fs)
        op_zones.append(op_fs)

    # scale zone hours such that they add up to 8736 
    # (if the surrogate is good, the unscaled will be pretty close to this)
    m.zone_total_hours = sum(op_zones[i].zone_hours for i in range(len(op_zones))) + off_fs.zone_hours

    for op_fs in op_zones:
        op_fs.scaled_zone_hours = Var(within=NonNegativeReals, bounds=(0,8736), initialize=100)
        # NOTE: scaled_hours_i = surrogate_i * 8736 / surrogate_total
        op_fs.con_scale_zone_hours = Constraint(
            expr = op_fs.scaled_zone_hours*m.zone_total_hours == op_fs.zone_hours * 8736
        )

    off_fs.scaled_zone_hours = Var(within=NonNegativeReals, bounds=(0, 8736), initialize=100)
    off_fs.con_scale_zone_hours = Constraint(
        expr = off_fs.scaled_zone_hours*m.zone_total_hours == off_fs.zone_hours*8736
    )

    #operating cost in $MM (million dollars)
    m.op_expr = sum(
        op_zones[i].scaled_zone_hours * op_zones[i].fs.operating_cost 
        for i in range(len(op_zones))
    ) * 1e-6 + off_fs.scaled_zone_hours * off_fs.fs.operating_cost * 1e-6

    #startup cost in MM$
    m.startup_expr = m.startup_cst * m.nstartups * m.pmax * 1e-6 #MM$

    #set zone flowsheets to pyomo model
    m.op_zones = op_zones

    #Piecewise cost limits, connect marginal cost to operating cost. We say marginal cost is the average operating cost
    m.connect_mrg_cost = Constraint(
        expr=m.marg_cst == 0.5*(op_zones[0].fs.operating_cost/m.pmin 
                                + op_zones[-1].fs.operating_cost/m.pmax)
    )

    # Expression for total cap and op cost - $
    m.total_cost = Expression(
        expr=plant_lifetime*(m.op_expr  + m.startup_expr) + capital_payment_years*cap_expr
    )

    # Expression for total revenue
    m.total_revenue = Expression(expr=plant_lifetime*m.revenue)

    # Objective $
    m.obj = Objective(expr=-(m.total_revenue - m.total_cost))

    # Unfixing the boiler inlet flowrate for capex plant
    m.cap_fs.fs.boiler.inlet.flow_mol[0].unfix()

    # Setting bounds for the capex plant flowrate
    m.cap_fs.fs.boiler.inlet.flow_mol[0].setlb(0.01)

    # Setting bounds for net cycle power output for the capex plant
    m.cap_fs.fs.eq_min_power = Constraint(
        expr=m.cap_fs.fs.net_cycle_power_output >= p_lower_bound*1e6)

    m.cap_fs.fs.eq_max_power = Constraint(
        expr=m.cap_fs.fs.net_cycle_power_output <= p_upper_bound*1e6)

    return m


if __name__ == "__main__":
    mdl = conceptual_design_problem_alamo(
        heat_recovery=True, 
        calc_boiler_eff=True,
    )

    solver = get_solver()
    solver.solve(mdl, tee=True)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

import numpy as np
import pandas as pd
import os
import pickle
import json
from idaes.core.surrogate.alamopy import AlamoTrainer
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Input
from tensorflow.keras.optimizers import Adam
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import cross_val_score
import tensorflow as tf
tf.keras.backend.set_floatx('float64')


def train_surrogate_alamo(
    x_train,
    z_train,
    x_labels,
    z_labels,
    config,
    file_dir="",
    filename="surrogate",
):
    # Determine the range of each input
    num_inputs = x_train.shape[1]
    x_min = list(x_train.min(axis=0))
    x_max = list(x_train.max(axis=0))

    # Store the data in a DataFrame
    data_in = pd.DataFrame(x_train, columns=x_labels)
    data_out = pd.DataFrame(z_train, columns=z_labels)
    data = pd.concat([data_in, data_out], axis=1)

    input_bounds = {x_labels[i]: (x_min[i], x_max[i]) for i in range(num_inputs)}
    config["filename"] = file_dir + filename + "_run.alm"
    config["overwrite_files"] = True

    trainer = AlamoTrainer(
        input_labels=x_labels,
        output_labels=z_labels,
        input_bounds=input_bounds,
        training_dataframe=data,
    )

    for opt, val in config.items():
        setattr(trainer.config, opt, val)

    # Train the surrogate model
    success, alm_surr, msg = trainer.train_surrogate()

    # Save the surrogate to file
    alm_surr.save_to_file(file_dir + filename + ".json", overwrite=True)

    return success, alm_surr, msg


def train_surrogate_keras(
    x_train,
    z_train,
    hidden_layers=(100, 50),
    activation="tanh",
    filename="surrogate",
):
    num_inputs = x_train.shape[1]
    num_outputs = (z_train.shape[1] if len(z_train.shape) == 2 else 1)

    model = Sequential(name=filename)
    model.add(Input(num_inputs))

    for nodes in hidden_layers:
        model.add(Dense(nodes, activation=activation))
    
    model.add(Dense(num_outputs))
    model.compile(optimizer=Adam(), loss='mse')
    history = model.fit(x=x_train, y=z_train, verbose=1, epochs=500)

    model.save("keras_models/" + filename)


def train_surrogate_scikit(
    x_train,
    z_train,
    activation="tanh",
    hidden_layers=(100, 50),
    filename="surrogate",
):
    # train scikit MLP Regressor model
    print("Training NN model ...")
    model = MLPRegressor(
        activation=activation,
        hidden_layer_sizes=hidden_layers,
    ).fit(x_train, z_train)

    # check the cross validation scores
    # print("Running Cross Validation Score...")
    # scores = cross_val_score(model, x_train, z_train, cv=5)

    # save model to pickle
    print("Saving model ...")
    with open("scikit_models/" + filename + ".pkl", 'wb') as f:
        pickle.dump(model, f)


def train_surrogate_model(
    x_data,
    z_data,
    x_labels=None,
    z_labels=None,
    trainer="alamo",
    shift_scale_data=True,
    alamo_config={"constant": True, "linfcns": True},
    activation="tanh",
    hidden_layers=(100, 50),
    filename="surrogate",
):
    """Build surrogate model for the given input-output data.

    Args:
        x_data: 
        z_data:
        x_label:
        z_label:
        trainer:
        shift_scale_data
    """

    if trainer not in ["alamo", "keras", "scikit"]:
        raise Exception(f"Unrecognized trainer {trainer}; Choose from [alamo, keras, scikit]")

    num_inputs = x_data.shape[1]
    num_outputs = (1 if len(z_data.shape) == 1 else z_data.shape[1])

    # Create default lables if labels are not provided
    if x_labels is None:
        x_labels = ["x" + str(i) for i in range(1, num_inputs + 1)]

    if z_labels is None:
        z_labels = ["z" + str(i) for i in range(1, num_outputs + 1)]

    # save scaling and training bounds
    xm = np.mean(x_data, axis=0)
    xstd = np.std(x_data, axis=0)
    zm = np.mean(z_data, axis=0)
    zstd = np.std(z_data, axis=0)
    xmin = list(np.min(x_data, axis=0))
    xmax = list(np.max(x_data, axis=0))
    data = {
        "xm_inputs": list(xm), 
        "xstd_inputs": list(xstd),
        "xmin": xmin,
        "xmax": xmax,
        "zm": list(zm) if hasattr(zm, "__iter__") else zm, 
        "zstd": list(zstd) if hasattr(zstd, "__iter__") else zstd,
    }

    cwd = os.getcwd()
    with open(cwd + "\\training_parameters_" + filename + ".json", 'w') as outfile:
        json.dump(data, outfile)

    # Shift the data by its mean value and scale it by its standard deviation
    if shift_scale_data:
        x_train_scaled = (x_data - xm) / xstd
        z_train_scaled = (z_data - zm) / zstd

    else:
        x_train_scaled = x_data
        z_train_scaled = z_data

    if trainer == "alamo":
        # Create a directory to store alamo models
        alm_dir = cwd + "\\alamo_models\\"
        if not os.path.exists(alm_dir):
            os.mkdir(os.path.join(cwd, "alamo_models"))

        success, alm_surr, msg = train_surrogate_alamo(
            x_train=x_train_scaled,
            z_train=z_train_scaled,
            x_labels=x_labels,
            z_labels=z_labels,
            config=alamo_config,
            file_dir=alm_dir,
            filename=filename,
        )

        return alm_surr

    elif trainer == "keras":
        # Create a directory to store keras models
        krs_dir = cwd + "\\keras_models\\"
        if not os.path.exists(krs_dir):
            os.mkdir(os.path.join(cwd, "keras_models"))

        train_surrogate_keras(
            x_train=x_train_scaled,
            z_train=z_train_scaled,
            activation=activation,
            hidden_layers=hidden_layers,
            filename=filename,
        )

    elif trainer == "scikit":
        # Create a directory to store keras models
        krs_dir = cwd + "\\scikit_models\\"
        if not os.path.exists(krs_dir):
            os.mkdir(os.path.join(cwd, "scikit_models"))

        train_surrogate_scikit(
            x_train=x_train_scaled,
            z_train=z_train_scaled,
            activation=activation,
            hidden_layers=hidden_layers,
            filename=filename,
        )



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
import numpy as np
import pandas as pd
from pathlib import Path
import utm
import networkx as nx


def read_prescient_file(filepath):
    df = pd.read_csv(filepath)
    if 'Minute' in df.columns:
        df['Datetime'] = pd.to_datetime(df['Date'].astype('str') + " " + df['Hour'].astype('str') + ":" + df['Minute'].astype('str'), format='%Y-%m-%d %H:%M')
        df = df.drop(['Date', 'Hour', 'Minute'], axis=1)
    else:
        df['Datetime'] = pd.to_datetime(df['Date'].astype('str') + " " + df['Hour'].astype('str') + ":00", format='%Y-%m-%d %H:%M')
        df = df.drop(['Date', 'Hour'], axis=1)
    
    df = df.set_index(pd.DatetimeIndex(df['Datetime']))
    df = df.drop(['Datetime'], axis=1)
    return df


def read_prescient_outputs(output_dir, source_dir, gen_name=None):
    """
    Read the bus information, bus LMPs, and renewables output detail CSVs from Prescient simulations

    Args:
        output_dir: Prescient simulation output directory
        source_dir: "SourceData" folder of the RTS-GMLC
    """
    source_dir = Path(source_dir)
    output_dir = Path(output_dir)

    summary = read_prescient_file(output_dir / "hourly_summary.csv")
    
    bus_detail_df = read_prescient_file(output_dir / "bus_detail.csv")
    bus_detail_df['LMP'] = bus_detail_df['LMP'].astype('float64')
    bus_detail_df['LMP DA'] = bus_detail_df['LMP DA'].astype('float64')
    bus_detail_df = bus_detail_df.rename(columns={'Demand': 'BusDemand', 'Overgeneration': 'BusOvergeneration'})

    summary = pd.merge(summary, bus_detail_df, how='outer', on=['Datetime'])

    renewables_df = read_prescient_file(output_dir / "renewables_detail.csv")
    thermal_df = read_prescient_file(output_dir / 'thermal_detail.csv')

    load_renewables = gen_name == None or gen_name in renewables_df.Generator.unique()
    load_thermal = gen_name == None or gen_name in thermal_df.Generator.unique()
    
    if load_renewables:
        gen_df = renewables_df

    if load_thermal:
        gen_df = thermal_df

    if load_thermal and load_renewables:
        gen_df = pd.merge(renewables_df, thermal_df, how='outer', on=['Datetime', 'Generator', 'Unit Market Revenue', 'Unit Uplift Payment'])

    return summary, gen_df


def read_rts_gmlc_wind_inputs_with_fix(source_dir, gen_df, aggfunc='first'):
    """
    For simulations that don't start at 01-01, the time index of the CF here may be shifted by one day relative to the output reported from renewables_detail.csv.
    To check that the time indices are correct, compare the DA output with the CF and make sure that the output never exceeds the CF.

    Use '317_WIND_1' to checkt he issue with the timestep. If the check passes with 317_wind_1 it should pass for the others as well.

    """
    gen_317_WIND_1='317_WIND_1'
    pmax_317_WIND_1 = 799.1

    gen_wind = gen_df[gen_df['Generator'] == gen_317_WIND_1]

    wind_df = read_rts_gmlc_wind_inputs(source_dir, None, aggfunc)

    if len(wind_df) == len(gen_wind):
        assert (wind_df.index == gen_wind.index).all()
        wind_cfs = wind_df
        # DA cfs should have no problems
        if not (wind_cfs[f'{gen_317_WIND_1}-RTCF'] * pmax_317_WIND_1 - gen_wind['Output']).min() > -1e-4:
            # try to fix by changing the agg func
            wind_cfs = read_rts_gmlc_wind_inputs(source_dir, gen_317_WIND_1, agg_func='mean')
    else:
        wind_cfs = wind_df[wind_df.index.isin(gen_wind.index)]
        if not (wind_cfs[f'{gen_317_WIND_1}-DACF'] * pmax_317_WIND_1 - gen_wind['Output DA']).min() > -1e-4:
            # try shifting the time index
            for c in wind_df.columns:
                wind_df[c] = np.roll(wind_df[c].values, 1)
            wind_cfs = wind_df[wind_df.index.isin(gen_wind.index)]
        if not (wind_cfs[f'{gen_317_WIND_1}-RTCF'] * pmax_317_WIND_1 - gen_wind['Output']).min() > -1e-4:
            # try to fix by changing the agg func
            if aggfunc != 'mean':
                wind_cfs = read_rts_gmlc_wind_inputs_with_fix(source_dir, gen_df, aggfunc='mean')

    assert (wind_cfs[f'{gen_317_WIND_1}-DACF'] * pmax_317_WIND_1 - gen_wind['Output DA']).min() > -1e-4
    assert (wind_cfs[f'{gen_317_WIND_1}-RTCF'] * pmax_317_WIND_1 - gen_wind['Output']).min() > -1e-4

    return wind_cfs
    

def read_rts_gmlc_wind_inputs(source_dir, gen_name=None, agg_func="first"):
    """
    Read capacity factors for day ahead and real time wind forecasts. 
    The forecasts are provided as 12 periods per hour, or 288 per day, and are aggregated to hourly data. 
    The aggregation can be done by taking the mean of the 12 periods, or taking the first. Check your Prescient parameters or outputs to verify.

    Warnings: 

    For simulations that don't start at 01-01, the time index of the CF here may be shifted by one day relative to the output reported from renewables_detail.csv.
    To check that the time indices are correct, compare the DA output with the CF and make sure that the output never exceeds the CF.

    Use `read_rts_gmlc_wind_inputs_with_fix` to check and try to fix these issues

    Args:
        source_dir: "SourceData" folder of the RTS-GMLC
        gen_name: optional wind generator name, if not provided then all wind generators returned
    """
    source_dir = Path(source_dir)
    gen_df = pd.read_csv(source_dir / "gen.csv")
    if not gen_name:
        wind_gens = [i for i in gen_df['GEN UID'] if "WIND" in i]
    else:
        wind_gens = [gen_name]

    wind_rt_df = pd.read_csv(source_dir.parent / "timeseries_data_files" / "WIND" / "REAL_TIME_wind.csv")
    wind_da_df = pd.read_csv(source_dir.parent / "timeseries_data_files" / "WIND" / "DAY_AHEAD_wind.csv")

    start_df = wind_rt_df.head(1)
    start_year = start_df.Year.values[0]
    start_mon = start_df.Month.values[0]
    start_day = start_df.Day.values[0]

    start_date = pd.Timestamp(f"{start_year}-{start_mon:02d}-{start_day:02d} 00:00:00")
    ix = pd.date_range(start=start_date, 
                        end=start_date
                        + pd.offsets.DateOffset(days=366)
                        - pd.offsets.DateOffset(hours=1),
                        freq='1H')
    wind_df = pd.DataFrame(index=ix)
    for k in wind_gens:
        rt_wind = wind_rt_df[k].values
        rt_wind = np.reshape(rt_wind, (8784, 12))
        if agg_func == "first":
            rt_wind = rt_wind[:, 0]
        elif agg_func == "mean":
            rt_wind = rt_wind.mean(1)
        else:
            raise ValueError(f"Unrecognized agg_func {agg_func}. Options are 'first' or 'mean'.")
        da_wind = wind_da_df[k].values

        wind_pmax = gen_df[gen_df['GEN UID'] == k]['PMax MW'].values[0]
        wind_df[k+"-RTCF"] = rt_wind / wind_pmax
        wind_df[k+"-DACF"] = da_wind / wind_pmax
    return wind_df


def get_rtsgmlc_bus_dict(source_dir):
    source_dir = Path(source_dir)
    bus_names = pd.read_csv(source_dir / "bus.csv")
    bus_dict = {k: v for k, v in zip(bus_names['Bus ID'].values, bus_names['Bus Name'].values)}
    return bus_dict


def prescient_outputs_for_gen(output_dir, source_dir, gen_name):
    """
    Get timeseries RT and DA Outputs and Capacity factors, Curtailment, Unit Market Revenue, Unit Uplift Payment for the given generator,
    and also the Demand, Shortfall, Overgeneration, and LMPs for the bus of the generator

    Args:
        output_dir: Prescient simulation output directory
        source_dir: "SourceData" folder of the RTS-GMLC
        gen_name: optional wind generator name, if not provided then all wind generators returned
    """
    source_dir = Path(source_dir)
    output_dir = Path(output_dir)
    summary, gen_df = read_prescient_outputs(output_dir, source_dir, gen_name)
    if "WIND" in gen_name or "PV" in gen_name:
        # double loop may have set the wind or pv plant as a thermal generator, so then these columns are not meaningful
        summary = summary.drop(['RenewablesUsed', 'RenewablesCurtailment'], axis=1)

    bus_dict = get_rtsgmlc_bus_dict(source_dir)
    bus_name = bus_dict[int(gen_name.split('_')[0])]

    summary = summary[summary.Bus == bus_name]
    gen_df = gen_df[gen_df.Generator == gen_name]
    df = pd.concat([summary, gen_df], axis=1)

    if 'WIND' in gen_name:
        wind_forecast_df = read_rts_gmlc_wind_inputs(source_dir, gen_name)
        wind_forecast_df = wind_forecast_df[wind_forecast_df.index.isin(df.index)]
        df = pd.concat([df, wind_forecast_df], axis=1)
    return df


def prescient_double_loop_outputs_for_gen(output_dir):
    """
    Get timeseries double-loop simulation outputs
    """
    output_dir = Path(output_dir)
    tracker_df = read_prescient_file(output_dir / "tracker_detail.csv")

    tracker_model_df = read_prescient_file(output_dir / "tracking_model_detail.csv")
    gen_name = tracker_model_df['Generator'].values[0]
    tracker_model_df = tracker_model_df.drop(['Generator'], axis=1)

    tracker_df = pd.merge(tracker_df, tracker_model_df, how='left', on=['Datetime', 'Horizon [hr]'])
    tracker_df.loc[:, 'Model'] = 'Tracker'

    bidder_df = pd.read_csv(output_dir / "bidder_detail.csv")
    da_bidder_df = bidder_df[bidder_df['Market'] == 'Day-ahead']
    da_bidder_df = da_bidder_df.rename(columns={'Hour': 'Horizon [hr]', 'Market': 'Model', 'Power 0 [MW]': 'DA Power 0 [MW]', 'Cost 0 [$]': 'DA Cost 0 [$]', 'Power 1 [MW]': 'DA Power 1 [MW]',
       'Cost 1 [$]': 'DA Cost 1 [$]', 'Power 2 [MW]': 'DA Power 2 [MW]', 'Cost 2 [$]': 'DA Cost 2 [$]'})
    da_bidder_df.loc[:, 'Model'] = 'DA Bidder'
    da_bidder_df['Datetime'] = pd.to_datetime(da_bidder_df['Date'].astype('str') + " 00:00:00", format='%Y-%m-%d %H:%M:%S')
    da_bidder_df = da_bidder_df.set_index(pd.DatetimeIndex(da_bidder_df['Datetime']))
    da_bidder_df = da_bidder_df.drop(['Date', 'Datetime', 'Generator'], axis=1)

    rt_bidder_df = bidder_df[bidder_df['Market'] == 'Real-time']
    rt_bidder_df = rt_bidder_df.rename(columns={'Market': 'Model', 'Power 0 [MW]': 'RT Power 0 [MW]', 'Cost 0 [$]': 'RT Cost 0 [$]', 'Power 1 [MW]': 'RT Power 1 [MW]',
       'Cost 1 [$]': 'RT Cost 1 [$]', 'Power 2 [MW]': 'RT Power 2 [MW]', 'Cost 2 [$]': 'RT Cost 2 [$]'})
    rt_bidder_df.loc[:, 'Model'] = 'RT Bidder'
    rt_bidder_df['Hour'] = np.tile(np.repeat(range(24), 4), 366)[0:len(rt_bidder_df)]
    rt_bidder_df['Horizon [hr]'] = np.tile(range(4), 8784)[0:len(rt_bidder_df)]
    rt_bidder_df['Datetime'] = pd.to_datetime(rt_bidder_df['Date'].astype('str') + " " + rt_bidder_df['Hour'].astype('str') + ":00", format='%Y-%m-%d %H:%M')
    rt_bidder_df = rt_bidder_df.set_index(pd.DatetimeIndex(rt_bidder_df['Datetime']))
    rt_bidder_df = rt_bidder_df.drop(['Date', 'Hour', 'Datetime', 'Generator'], axis=1)

    bidder_df = pd.merge(da_bidder_df, rt_bidder_df, how='outer', on=['Datetime', 'Horizon [hr]', 'Model'])
    df = pd.merge(bidder_df, tracker_df, how='outer', on=['Datetime', 'Horizon [hr]', 'Model'])
    return df, gen_name


def double_loop_outputs_for_gen(double_loop_dir, source_dir):
    source_dir = Path(source_dir)
    double_loop_dir = Path(double_loop_dir)
    dl_df, gen_name = prescient_double_loop_outputs_for_gen(double_loop_dir)
    res_df = prescient_outputs_for_gen(double_loop_dir, source_dir, gen_name)
    res_df = res_df[res_df.index.isin(dl_df.index.unique())]
    res_df['Datetime'] = res_df.index
    res_df.loc[:, 'Model'] = "Prescient"

    df = pd.merge(dl_df, res_df, how='outer', on=['Datetime', 'Model']).set_index('Datetime')
    df.Model = pd.Categorical(df.Model, categories=['DA Bidder', 'RT Bidder', 'Tracker', "Prescient"])
    df = df.sort_values(by = ['Datetime', 'Model'], na_position = 'last')
    return df.dropna(axis=1, how='all')


def get_rtsgmlc_network(output_dir, source_dir, length_scaling=1/300, width_scaling=1/75):
    edges_df = pd.read_csv(source_dir / "branch.csv").set_index("UID")
    nodes_df = pd.read_csv(source_dir / "bus.csv").set_index("Bus ID")
    nodes_df.index = nodes_df.index.map(str)
    line_detail = read_prescient_file(output_dir / "line_detail.csv")
    summary, gen_df = read_prescient_outputs(output_dir, source_dir)
    bus_dict = {k: v for k, v in zip(nodes_df.index.values, nodes_df['Bus Name'].values)}

    # Create Network with Edge info
    cont_rating = edges_df['Cont Rating'].values.tolist()
    line_detail['Cont Rating'] = cont_rating * len(line_detail.index.unique())
    line_detail["Relative Flow"] = np.abs(line_detail['Flow'] / line_detail['Cont Rating'])

    G = nx.from_pandas_edgelist(edges_df.reset_index().astype('str'), source='From Bus', target='To Bus', edge_attr=('Cont Rating', 'LTE Rating', 'STE Rating', 'Length', "UID"))

    # Fill in Node X, Y info
    nodes_df['x'], nodes_df['y'] = utm.from_latlon(nodes_df['lat'].values, nodes_df['lng'].values)[0:2]

    x_pos = (nodes_df['x'] - nodes_df['x'].mean()) * length_scaling
    y_pos = (nodes_df['y'] - nodes_df['y'].mean()) * length_scaling
    nx.set_node_attributes(G, x_pos.to_dict(), name="x")
    nx.set_node_attributes(G, y_pos.to_dict(), name="y")
    nx.set_node_attributes(G, nodes_df['Bus Name'].to_dict(), name="name")

    # Fill in Node colors by generators
    color_dict = {}
    wind_buses = gen_df[gen_df['Generator'].str.contains('WIND')]['Generator'].str.slice(0, 3).unique()
    pv_buses = gen_df[gen_df['Generator'].str.contains('PV')]['Generator'].str.slice(0, 3).unique()
    nuc_buses = gen_df[gen_df['Generator'].str.contains('NUCLEAR')]['Generator'].str.slice(0, 3).unique()

    for n in G.nodes:
        if n in wind_buses and n in pv_buses and n in nuc_buses:
            color = 'white'
        elif n in wind_buses and n in pv_buses:
            color = 'green'
        elif n in wind_buses and n in nuc_buses:
            color = "purple"
        elif n in pv_buses and n in nuc_buses:
            color = 'orange'
        elif n in wind_buses:
            color = 'blue'
        elif n in pv_buses:
            color = 'yellow'
        elif n in nuc_buses:
            color = "red"
        else:
            color = 'grey'
        color_dict[n] = color
    nx.set_node_attributes(G, color_dict, name="color")

    # Add Edge Width for Rating and congestion for color
    line_detail['Relative OverFlow'] = (line_detail['Relative Flow'] >= 0.95).astype(int) * line_detail['Relative Flow']
    edges_df['Relative OverFlow'] = line_detail.set_index('Line')['Relative OverFlow'].groupby('Line').mean()
    edges_df['Relative Flow'] = line_detail.set_index('Line')['Relative Flow'].groupby('Line').mean()
    edges_df

    avg_LMP_diff = []
    for (i, j) in edges_df[['From Bus', 'To Bus']].values:
        LMP_diff = (summary.query(f"Bus == '{bus_dict[str(i)]}'")['LMP'] - summary.query(f"Bus == '{bus_dict[str(j)]}'")['LMP']).abs()
        avg_LMP_diff.append(LMP_diff.mean())
    edges_df['Avg LMP Diff'] = avg_LMP_diff
    edges_df

    edges_dict = {}
    edge_color_val = 'Avg LMP Diff' # or 'Relative OverFlow'
    max_cong = edges_df[edge_color_val].max()
    for (i, j, w, o) in edges_df[['From Bus', 'To Bus', 'Cont Rating', edge_color_val]].values:
        if o != 0:
            rel_cong = o / max_cong
            gb =  75 + int(180 * rel_cong)
            color = '#%02x%02x%02x' % (gb, 0, 0)
        else:
            color = 'black'
    for (i, j, w, o) in edges_df[['From Bus', 'To Bus', 'Cont Rating', edge_color_val]].values:
        if o != 0:
            rel_cong = o / max_cong
            gb =  75 + int(180 * rel_cong)
            color = '#%02x%02x%02x' % (gb, 0, 0)
        else:
            color = 'black'
        edges_dict[(str(int(i)), str(int(j)))] = {'width': w * width_scaling, 'color': color, 'title': str(round(o, 2)), edge_color_val: o}
    nx.set_edge_attributes(G, edges_dict)
    return G, line_detail, edges_df, gen_df, summary


#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Simple rankine cycle model. Has couple of options:
1. Recover waste heat after turbine to mimic feed water heater integration
2. Option to include boiler efficiency which is a linear fit f(capacity factor)

if no heat recovery, the flowsheet is as follows:
    Boiler --> Turbine --> Condenser --> Pump --> Boiler

if heat_recovery, the flowsheet is as follows:
    Boiler --> Turbine --> pre-condenser(- Q_recovered) --> Condenser -->
    Pump --> Feed water heater(+ Q_recovered) --> Boiler

Note:
* Boiler and condenser are simple heater blocks
* IAPWS95 for water and steam properties
"""

__author__ = "Jaffer Ghouse"


# Import Pyomo libraries
from pyomo.environ import (
    ConcreteModel, units, Var, TransformationFactory, value, Block, 
    Expression, Constraint, Param, Objective, units as pyunits,
)
from pyomo.network import Arc

# Import IDAES components
from idaes.core import FlowsheetBlock, UnitModelBlockData, UnitModelCostingBlock

# Import heat exchanger unit model
from idaes.models.unit_models import Heater, PressureChanger

from idaes.models.unit_models.pressure_changer import ThermodynamicAssumption
from idaes.models_extra.power_generation.costing.power_plant_capcost \
    import QGESSCosting, QGESSCostingData

# Import steam property package
from idaes.models.properties.iapws95 import htpx, Iapws95ParameterBlock

from idaes.apps.grid_integration import MultiPeriodModel
from idaes.core.util import to_json, from_json
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import propagate_state
from idaes.core.solvers import get_solver

# Import IDAES logger
import idaes.logger as idaeslog


def create_model(
    m=None,
    heat_recovery=False, 
    calc_boiler_eff=False, 
    capital_fs=False,
):

    if m is None:
        m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.steam_prop = Iapws95ParameterBlock()

    m.fs.boiler = Heater(
        property_package=m.fs.steam_prop,
        has_pressure_change=False,
    )

    m.fs.turbine = PressureChanger(
        property_package=m.fs.steam_prop,
        compressor=False,
        thermodynamic_assumption=ThermodynamicAssumption.isentropic,
    )

    if heat_recovery:
        m.fs.pre_condenser = Heater(
            property_package=m.fs.steam_prop,
            has_pressure_change=True,
        )

        # Spec for pre-condenser
        m.fs.pre_condenser.eq_outlet_cond = Constraint(
            expr=m.fs.pre_condenser.control_volume.
            properties_out[0].enth_mol == m.fs.pre_condenser.control_volume.
            properties_out[0].enth_mol_sat_phase["Liq"]
        )

        m.fs.feed_water_heater = Heater(
            property_package=m.fs.steam_prop,
            has_pressure_change=True,
        )

        # Link precondenser heat and feed water heater
        m.fs.eq_heat_recovery = Constraint(
            expr=m.fs.pre_condenser.heat_duty[0] ==
            - m.fs.feed_water_heater.heat_duty[0]
        )

    m.fs.condenser = Heater(
        property_package=m.fs.steam_prop,
        has_pressure_change=True,
    )

    m.fs.bfw_pump = PressureChanger(
        property_package=m.fs.steam_prop,
        thermodynamic_assumption=ThermodynamicAssumption.pump,
    )

    # create arcs
    m.fs.boiler_to_turbine = Arc(source=m.fs.boiler.outlet,
                                 destination=m.fs.turbine.inlet)

    if heat_recovery:
        m.fs.turbine_to_precondenser = Arc(
            source=m.fs.turbine.outlet,
            destination=m.fs.pre_condenser.inlet)
        m.fs.precondenser_to_condenser = Arc(
            source=m.fs.pre_condenser.outlet,
            destination=m.fs.condenser.inlet)
        m.fs.pump_to_feedwaterheater = Arc(
            source=m.fs.bfw_pump.outlet,
            destination=m.fs.feed_water_heater.inlet)
    else:
        m.fs.turbine_to_condenser = Arc(source=m.fs.turbine.outlet,
                                        destination=m.fs.condenser.inlet)

    m.fs.condenser_to_pump = Arc(source=m.fs.condenser.outlet,
                                 destination=m.fs.bfw_pump.inlet)

    # expand arcs
    TransformationFactory("network.expand_arcs").apply_to(m)

    # Compute gross power
    m.fs.gross_cycle_power_output = Expression(
        expr=(-m.fs.turbine.work_mechanical[0] - m.fs.bfw_pump.work_mechanical[0])
    )

    # account for generator loss = 5% of gross power output
    m.fs.net_cycle_power_output = Expression(
        expr=0.95 * m.fs.gross_cycle_power_output
    )

    if capital_fs or not calc_boiler_eff:
        # if fs is a capital cost fs, then the P is at P_max and hence
        # set boiler efficiency to value at P_max instead of computing
        m.fs.boiler_eff = Param(initialize=0.95)

    if calc_boiler_eff:

        # Var for net_power max variable.
        # This is needed to compute boiler efficiency as a function of
        # capacity factor; p and p_max must be in MWs
        m.fs.net_power_max = Var(initialize=100, units=pyunits.MW)

        # # Boiler efficiency
        # # Linear fit as function of capacity factor; at P_max eff. is 95%
        m.fs.boiler_eff = Expression(
            expr=0.2143 * (pyunits.convert(m.fs.net_cycle_power_output, pyunits.MW) / m.fs.net_power_max)
            + 0.7357
        )

    #  cycle efficiency
    m.fs.cycle_efficiency = Expression(
        expr=m.fs.net_cycle_power_output/m.fs.boiler.heat_duty[0]
        * m.fs.boiler_eff * 100
    )

    m.heat_recovery = heat_recovery
    m.calc_boiler_eff = calc_boiler_eff

    return m


def initialize_model(m, outlvl=idaeslog.INFO):

    # Deactivate the constraint linking the pre_condenser Q and feed water Q
    if m.heat_recovery:
        m.fs.eq_heat_recovery.deactivate()

    # Check for degrees of freedom before proceeding with initialization
    assert degrees_of_freedom(m) == 0

    # Proceed with initialization
    m.fs.boiler.initialize(outlvl=outlvl)
    propagate_state(m.fs.boiler_to_turbine)

    m.fs.turbine.initialize(outlvl=outlvl)

    if m.heat_recovery:
        propagate_state(m.fs.turbine_to_precondenser)
        m.fs.pre_condenser.initialize(outlvl=outlvl)

        propagate_state(m.fs.precondenser_to_condenser)
        m.fs.condenser.initialize()

        propagate_state(m.fs.condenser_to_pump)
        m.fs.bfw_pump.initialize(outlvl=outlvl)

        propagate_state(m.fs.pump_to_feedwaterheater)
        m.fs.feed_water_heater.initialize(outlvl=outlvl)
    else:
        propagate_state(m.fs.turbine_to_condenser)
        m.fs.condenser.initialize(outlvl=outlvl)

        propagate_state(m.fs.condenser_to_pump)
        m.fs.bfw_pump.initialize(outlvl=outlvl)

    solver = get_solver()
    solver.solve(m, tee=False)

    if m.heat_recovery:
        # Unfix feed water heater temperature as the constraint linking Q
        # will be activated
        m.fs.feed_water_heater.outlet.enth_mol[0].unfix()

        m.fs.eq_heat_recovery.activate()

    assert degrees_of_freedom(m) == 0
    solver.solve(m, tee=True)

    return m


def generate_report(m, unit_model_report=True):

    # Print reports
    if unit_model_report:
        for i in m.fs.component_objects(Block):
            if isinstance(i, UnitModelBlockData):
                i.report()

    print()
    print('Net power = ', value(m.fs.net_cycle_power_output)*1e-6, ' MW')
    print('Cycle efficiency = ', value(m.fs.cycle_efficiency), "%")
    print('Heat rate = ', value(m.fs.heat_rate), 'Btu/kWh')
    print('Boiler feed water flow = ',
          value(m.fs.boiler.inlet.flow_mol[0]), "mol/s")
    print()
    try:
        print('Capital cost = ', value(m.fs.capital_cost), '$M')
    except AttributeError:
        print("No cap cost for opex plant")
    try:
        print('Operating cost =  ',
              value(m.fs.operating_cost/(m.fs.net_cycle_power_output*1e-6)),
              '$/MWh')
    except AttributeError:
        print("No operating cost for capex plant")


def set_inputs(m, bfw_pressure=24.23e6, bfw_flow=10000):
    # Boiler inlet
    m.fs.boiler.inlet.flow_mol[0].fix(bfw_flow)  # mol/s
    m.fs.boiler.inlet.pressure[0].fix(bfw_pressure)  # MPa
    m.fs.boiler.inlet.enth_mol[0].fix(
        htpx(T=563.6*units.K, P=value(m.fs.boiler.inlet.pressure[0])*units.Pa)
    )

    # Unit specifications
    m.fs.boiler.outlet.enth_mol[0].fix(
        htpx(T=866.5*units.K, P=value(m.fs.boiler.inlet.pressure[0])*units.Pa)
    )

    turbine_pressure_ratio = 2e6 / bfw_pressure
    m.fs.turbine.ratioP.fix(turbine_pressure_ratio)
    m.fs.turbine.efficiency_isentropic.fix(0.85)

    if m.heat_recovery:
        # precondenser
        m.fs.pre_condenser.deltaP.fix(-0.5e6)  # Pa

        # feed water heater
        m.fs.feed_water_heater.deltaP[0].fix(0)  # Pa
        m.fs.feed_water_heater.outlet.enth_mol[0].fix(
            htpx(T=563.6*units.K, P=value(m.fs.condenser.outlet.pressure[0])*units.Pa)
        )

    m.fs.condenser.outlet.pressure[0].fix(1.05e6)  # Pa
    m.fs.condenser.outlet.enth_mol[0].fix(
        htpx(T=311*units.K, P=value(m.fs.condenser.outlet.pressure[0])*units.Pa)
    )

    m.fs.bfw_pump.efficiency_pump.fix(0.80)
    m.fs.bfw_pump.deltaP.fix(bfw_pressure)

    return m


def close_flowsheet_loop(m):

    """Closes the loop i.e. the arc between the feed water heater and
    boiler. When the pressure and enthalpy arcs are enabled, the bfw_pump
    spec for deltaP and the inlet enth_mol for the boiler need to be unfixed.

    Returns:
        m: model object after closing the loop
    """
    # Unfix bfw pump pressure spec
    m.fs.bfw_pump.deltaP.unfix()

    # Unfix inlet boiler enthalpy
    m.fs.boiler.inlet.enth_mol[0].unfix()

    if m.heat_recovery:
        # Constraint to link pressure
        m.fs.eq_pressure = Constraint(
            expr=m.fs.feed_water_heater.outlet.pressure[0] ==
            m.fs.boiler.inlet.pressure[0]
        )

        # Constraint to link enthalpy
        m.fs.eq_enthalpy = Constraint(
            expr=m.fs.feed_water_heater.outlet.enth_mol[0] ==
            m.fs.boiler.inlet.enth_mol[0]
        )
    else:
        # Constraint to link pressure
        m.fs.eq_pressure = Constraint(
            expr=m.fs.bfw_pump.outlet.pressure[0] ==
            m.fs.boiler.inlet.pressure[0]
        )

        # Constraint to link enthalpy
        m.fs.eq_enthalpy = Constraint(
            expr=m.fs.bfw_pump.outlet.enth_mol[0] ==
            m.fs.boiler.inlet.enth_mol[0]
        )

    return m


def add_capital_cost(m):

    """Add capital cost expressions. Leverages costing correlations from the
    IDAES costing library. Note that the capital cost correlations are all
    based on the boiler feed water flowrate.

    Returns:
        m: model object after adding capital cost correlations
    """
    m.fs.costing = QGESSCosting()

    # Add boiler capital cost
    boiler_power_account = ['4.9']
    # convert flow rate of BFW from mol/s to lb/hr for costing expressions
    m.fs.bfw_lb_hr = Expression(
        expr=pyunits.convert(
            m.fs.boiler.inlet.flow_mol[0] * m.fs.steam_prop.mw, pyunits.lb / pyunits.hr
        )
    )

    m.fs.boiler.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": boiler_power_account,
            "scaled_param": m.fs.bfw_lb_hr,
            "tech": 2,
        },
    )

    # Add turbine capital cost
    turb_power_account = ['8.1']
    # convert the turbine power from W to kW for costing expressions
    m.fs.turbine_power_mw = Expression(
        expr=pyunits.convert(-m.fs.turbine.work_mechanical[0], pyunits.kW)
    )

    m.fs.turbine.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": turb_power_account,
            "scaled_param": m.fs.turbine_power_mw,
            "tech": 2,
        }
    )

    # Add condenser cost
    cond_power_account = ['8.3']
    # convert the heat duty from J/s to MMBtu/hr for costing expressions
    m.fs.condenser_duty_mmbtu_h = Expression(
        expr=pyunits.convert(-m.fs.condenser.heat_duty[0], pyunits.MBtu / pyunits.hr),
    )

    m.fs.condenser.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": cond_power_account,
            "scaled_param": m.fs.condenser_duty_mmbtu_h,
            "tech": 2,
        }
    )

    # Add feed water system costs
    # Note that though no feed water heaters were used, BFW flowrate is used
    # to cost the fed water system
    fwh_power_account = ['3.1', '3.3', '3.5']

    m.fs.bfw_pump.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": fwh_power_account,
            "scaled_param": m.fs.bfw_lb_hr,
            "tech": 2,
        }
    )

    # Add expression for total capital cost
    m.fs.capital_cost = Expression(
        expr=m.fs.boiler.costing.total_plant_cost['4.9'] +
        m.fs.turbine.costing.total_plant_cost['8.1'] +
        m.fs.condenser.costing.total_plant_cost['8.3'] +
        sum(m.fs.bfw_pump.costing.total_plant_cost[:]),
        doc="Total capital cost $ Million")

    return m


def add_operating_cost(m, include_cooling_cost=True, coal_price=51.96):

    """Add operating cost expressions. The operating cost only includes
    the cost of coal. This is computed by calculating the amount of coal
    required based on HHV value of coal and the boiler heat duty.

    Returns:
        m: model object after adding operating cost correlations
    """

    # Add condenser cooling water cost
    # temperature for the cooling water from/to cooling tower in K
    t_cw_in = 289.15
    t_cw_out = 300.15

    # compute the delta_h based on fixed temperature of cooling water
    # utility
    m.fs.enth_cw_in = Param(
        initialize=htpx(T=t_cw_in*units.K, P=101325*units.Pa),
        doc="inlet enthalpy of cooling water to condenser",
        units=pyunits.J / pyunits.mol,
    )
    m.fs.enth_cw_out = Param(
        initialize=htpx(T=t_cw_out*units.K, P=101325*units.Pa),
        doc="outlet enthalpy of cooling water from condenser",
        units=pyunits.J / pyunits.mol,
    )
    m.fs.density_cw = Param(
        initialize=1000,
        doc="Density of cooling water in kg/m^3",
        units=pyunits.kg / pyunits.m ** 3,
    )

    m.fs.cw_flow = Expression(
        expr=pyunits.convert(
            -m.fs.condenser.heat_duty[0] /
            (m.fs.density_cw * (m.fs.enth_cw_out - m.fs.enth_cw_in) / m.fs.steam_prop.mw),
            pyunits.gal / pyunits.hr
        ),
        doc="cooling water flow rate in gallons/hr",
    )

    # cooling water cost is $0.19 per 1000 gallons
    m.fs.cw_cost = Param(
        initialize=0.19 / 1000,
        doc="cost of cooling water for condenser in $/gallon",
        units=pyunits.USD_2018 / pyunits.gal,
    )

    m.fs.cw_total_cost = Expression(
        expr=m.fs.cw_flow * m.fs.cw_cost,
        doc="total cooling water cost in $/hr",
    )

    # Add coal feed costs
    # HHV value of coal (Reference - NETL baseline report rev #4)
    m.fs.coal_hhv = Param(
        initialize=27113,
        doc="Higher heating value of coal as received kJ/kg",
        units=pyunits.kJ / pyunits.kg,
    )

    # cost of coal (Reference - NETL baseline report rev #4)
    m.fs.coal_cost = Param(
        initialize=coal_price,
        doc="$ per ton of Illinois no. 6 coal",
        units=pyunits.USD_2018 / pyunits.ton,
    )

    # Expression to compute coal flow rate in ton/hr using Q_boiler and
    # hhv values
    m.fs.coal_flow = Expression(
        expr=pyunits.convert(
            (m.fs.boiler.heat_duty[0] / m.fs.boiler_eff) 
            / pyunits.convert(m.fs.coal_hhv, pyunits.J / pyunits.kg),
            pyunits.ton / pyunits.hr
        ),
        doc="coal flow rate for boiler ton/hr",
    )
    # Expression to compute total cost of coal feed in $/hr
    m.fs.total_coal_cost = Expression(
        expr=m.fs.coal_flow * m.fs.coal_cost,
        doc="total cost of coal feed in $/hr"
    )

    # Expression to compute heat rate (Btu/kWh)
    # Factors:
    # 907.18 to convert from ton to Kg
    # 0.9478 to convert 1 KJ to 1 BTU
    # 1e3 to convert power in MW to kW

    m.fs.heat_rate = Expression(
        expr=pyunits.convert(m.fs.coal_flow * m.fs.coal_hhv, pyunits.Btu / pyunits.hr)
        / pyunits.convert(m.fs.net_cycle_power_output, pyunits.kW),
        doc="heat rate of plant in Btu/kWh",
    )

    if include_cooling_cost:
        # Expression for total operating cost
        m.fs.operating_cost = Expression(
            expr=m.fs.total_coal_cost + m.fs.cw_total_cost,
            doc="Total operating cost in $/hr",
        )
    else:
        # Expression for total operating cost
        m.fs.operating_cost = Expression(
            expr=m.fs.total_coal_cost,
            doc="Total operating cost in $/hr",
        )

    return m


def square_problem(
    heat_recovery=None,
    capital_fs=False,
    net_power=100,
    p_max=100,
    calc_boiler_eff=False,
    capital_payment_years=5,
):

    """This method simulates the simple rankine cycle by adding capital and
    operating costs.

    """
    # Create plant flowsheet
    m = create_model(
        heat_recovery=heat_recovery,
        capital_fs=capital_fs,
        calc_boiler_eff=calc_boiler_eff)

    # Set model inputs for the capex and opex plant
    set_inputs(m)

    # Set p_max for plant that is set in a square problem
    if calc_boiler_eff:
        m.fs.net_power_max.fix(p_max)

    # Initialize the capex and opex plant
    initialize_model(m)

    # Closing the loop in the flowsheet
    close_flowsheet_loop(m)

    # Unfixing the boiler inlet flowrate
    m.fs.boiler.inlet.flow_mol[0].unfix()

    # Net power constraint for the capex plant
    m.fs.eq_net_power = Constraint(
        expr=m.fs.net_cycle_power_output == net_power*1e6
    )

    add_capital_cost(m)

    add_operating_cost(m, include_cooling_cost=True)

    # Expression for total cap and op cost - $/hr
    m.total_cost = Expression(
        expr=(m.fs.capital_cost*1e6/capital_payment_years/8760) +
        m.fs.operating_cost)

    solver = get_solver()
    solver.solve(m, tee=True)

    generate_report(m, unit_model_report=False)

    return m


def stochastic_optimization_problem(
    heat_recovery=False,
    calc_boiler_eff=False,
    p_max_lower_bound=10,
    p_max_upper_bound=300,
    capital_payment_years=5,
    plant_lifetime=20,
    power_demand=None, 
    lmp=None,
    lmp_weights=None,
):

    """This method sets up the stochastic optimization problem that sets up a
    steady-state, pricetaker problem.

    Returns:
        [type]: [description]
    """
    m = ConcreteModel()

    # Create capex plant
    m.cap_fs = create_model(
        heat_recovery=heat_recovery,
        capital_fs=True, 
        calc_boiler_eff=False,
    )
    set_inputs(m.cap_fs)
    initialize_model(m.cap_fs)
    close_flowsheet_loop(m.cap_fs)
    add_capital_cost(m.cap_fs)

    # capital cost (M$/yr)
    cap_expr = pyunits.convert(
        m.cap_fs.fs.capital_cost / capital_payment_years, pyunits.USD_2018
    )

    # Create opex plant
    op_expr = 0
    rev_expr = 0

    for i in range(len(lmp)):

        print()
        print("Creating instance ", i)
        if not calc_boiler_eff:
            op_fs = create_model(
                heat_recovery=heat_recovery,
                capital_fs=False,
                calc_boiler_eff=False)

            # Set model inputs for the capex and opex plant
            set_inputs(op_fs)

            if i == 0:
                # Initialize the capex and opex plant
                initialize_model(op_fs)

                # save model state after initializing the first instance
                to_json(op_fs.fs, fname="initialized_state.json.gz",
                        gz=True, human_read=True)
            else:
                # Initialize the capex and opex plant
                from_json(op_fs.fs, fname="initialized_state.json.gz", gz=True)

            # Closing the loop in the flowsheet
            close_flowsheet_loop(op_fs)

            add_operating_cost(op_fs)

            op_expr += lmp_weights[i]*op_fs.fs.operating_cost
            rev_expr += lmp_weights[i]*lmp[i]*op_fs.fs.net_cycle_power_output*1e-6

            # Add inequality constraint linking net power to cap_ex
            # operating P_min <= 30% of design P_max
            op_fs.fs.eq_min_power = Constraint(
                expr=op_fs.fs.net_cycle_power_output >=
                0.3*m.cap_fs.fs.net_cycle_power_output)
            # operating P_max = design P_max
            op_fs.fs.eq_max_power = Constraint(
                expr=op_fs.fs.net_cycle_power_output <=
                m.cap_fs.fs.net_cycle_power_output)
        else:
            op_fs = create_model(
                heat_recovery=heat_recovery,
                capital_fs=False,
                calc_boiler_eff=True)

            # Set model inputs for the capex and opex plant
            set_inputs(op_fs)

            # Fix the p_max of op_fs to p of cap_fs for initialization
            op_fs.fs.net_power_max.fix(
                value(m.cap_fs.fs.net_cycle_power_output))

            if i == 0:
                # Initialize the capex and opex plant
                initialize_model(op_fs)

                # save model state after initializing the first instance
                to_json(op_fs.fs, fname="initialized_state.json.gz",
                        gz=True, human_read=True)
            else:
                # Initialize the capex and opex plant
                from_json(op_fs.fs, fname="initialized_state.json.gz", gz=True)

            # Closing the loop in the flowsheet
            close_flowsheet_loop(op_fs)
            add_operating_cost(op_fs)

            op_expr += lmp_weights[i]*op_fs.fs.operating_cost
            rev_expr += lmp_weights[i]*lmp[i]*op_fs.\
                fs.net_cycle_power_output*1e-6

            # Unfix op_fs p_max and set constraint linking that to cap_fs p_max
            op_fs.fs.net_power_max.unfix()
            op_fs.fs.eq_p_max = Constraint(
                expr=op_fs.fs.net_power_max ==
                m.cap_fs.fs.net_cycle_power_output*1e-6
            )

            # Add inequality constraint linking net power to cap_ex
            # operating P_min <= 30% of design P_max
            op_fs.fs.eq_min_power = Constraint(
                expr=op_fs.fs.net_cycle_power_output >=
                0.3*m.cap_fs.fs.net_cycle_power_output)
            # operating P_max = design P_max
            op_fs.fs.eq_max_power = Constraint(
                expr=op_fs.fs.net_cycle_power_output <=
                m.cap_fs.fs.net_cycle_power_output)

        # only if power demand is given
        if power_demand is not None:
            op_fs.fs.eq_max_produced = Constraint(
                expr=op_fs.fs.net_cycle_power_output <=
                power_demand[i]*1e6)

        op_fs.fs.boiler.inlet.flow_mol[0].unfix()

        # Set bounds for the flow
        op_fs.fs.boiler.inlet.flow_mol[0].setlb(1)
        # op_fs.fs.boiler.inlet.flow_mol[0].setub(25000)

        setattr(m, 'scenario_{}'.format(i), op_fs)

    # Expression for total cap and op cost - $
    m.total_cost = Expression(
        expr=plant_lifetime*op_expr + capital_payment_years*cap_expr)

    # Expression for total revenue
    m.total_revenue = Expression(
        expr=plant_lifetime*rev_expr)

    # Objective $
    m.obj = Objective(
        expr=-(m.total_revenue - m.total_cost))

    # Unfixing the boiler inlet flowrate for capex plant
    m.cap_fs.fs.boiler.inlet.flow_mol[0].unfix()

    # Setting bounds for the capex plant flowrate
    m.cap_fs.fs.boiler.inlet.flow_mol[0].setlb(5)

    # Setting bounds for net cycle power output for the capex plant
    m.cap_fs.fs.eq_min_power = Constraint(
        expr=m.cap_fs.fs.net_cycle_power_output >= p_max_lower_bound*1e6)

    m.cap_fs.fs.eq_max_power = Constraint(
        expr=m.cap_fs.fs.net_cycle_power_output <=
        p_max_upper_bound*1e6)

    return m


def unfix_dof_function(m, cap_fs):
    
    # Unfix inlet flowrate to boiler
    m.fs.boiler.inlet.flow_mol[0].unfix()

    # Set bounds for the flow
    m.fs.boiler.inlet.flow_mol[0].setlb(1)

    if m.calc_boiler_eff:
        m.fs.net_power_max.unfix()

        m.fs.eq_p_max = Constraint(
            expr=m.fs.net_power_max == cap_fs.fs.net_cycle_power_output * 1e-6
        )

    close_flowsheet_loop(m)


if __name__ == "__main__":

    # Code to generate op cost, heat rate, eff vs. capacity factor plot
    """
    p_max = 300
    p_min = 90
    power = list(reversed(range(p_min, p_max + 30, 30)))
    plant_capacity = [p*100/p_max for p in power]
    cycle_eff = []
    heat_rate = []
    op_cost = []
    for i in power:
        print(i)
        m = square_problem(
            heat_recovery=True,
            capital_fs=False,
            calc_boiler_eff=True,
            p_max=p_max, net_power=i)
        cycle_eff.append(value(m.fs.cycle_efficiency))
        heat_rate.append(value(m.fs.heat_rate))
        op_cost.append(value(m.fs.operating_cost)/i)

    # plots
    from matplotlib import pyplot as plt
    fig, ax = plt.subplots()
    plt.plot(plant_capacity, op_cost, color="green")
    sec_yaxis = ax.secondary_yaxis('left')

    ax.set_xlabel("operating capacity (%)")
    ax.set_ylabel("operating cost ($/MWh)", color="green")

    ax1 = ax.twinx()
    ax1.plot(plant_capacity, heat_rate, color="green")
    ax1.yaxis.set_ticks_position('left')
    ax1.yaxis.set_label_position('left')
    ax1.spines['left'].set_position(('outward', 75))
    ax1.set_ylabel("heat rate (BTU/kWh)", color="green")

    ax2 = ax.twinx()
    ax2.plot(plant_capacity, cycle_eff, color="red")
    ax2.set_ylabel("cycle efficiency (%)", color="red")
    plt.grid()
    # plt.savefig("operating_cost_vs_plant_capacity.pdf",
    #             format="pdf",
    #             bbox_inches="tight")
    plt.show()
    """

    # Code to generate capital cost as f(P_max)
    m = square_problem(
        heat_recovery=True,
        capital_fs=True,
        calc_boiler_eff=True,
        p_max=300, net_power=300)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
from prescient.simulator import Prescient
from types import ModuleType
from argparse import ArgumentParser
import pandas as pd
from pathlib import Path
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
from idaes.apps.grid_integration import Tracker
from idaes.apps.grid_integration.model_data import ThermalGeneratorModelData
from dispatches.workflow.coordinator import DoubleLoopCoordinator
from dispatches_sample_data import rts_gmlc
from dispatches.case_studies.renewables_case.wind_PEM_double_loop import MultiPeriodWindPEM
from dispatches.case_studies.renewables_case.load_parameters import *
from dispatches.case_studies.renewables_case.double_loop_utils import read_rts_gmlc_wind_inputs
from dispatches.case_studies.renewables_case.PEM_parametrized_bidder import PEMParametrizedBidder, PerfectForecaster

###
# Script to run a double loop simulation of a Wind + PEM flowsheet (MultiPeriodWindPEM) in Prescient.
# Plant's system dynamics and costs are in wind_PEM_double_loop.
# Plant's bid parameter is "pem_bid" and is constant throughout the simulation.
# Bid curve is a constant function of the available wind resource (one per DA or RT).
# 
# The DoubleLoopCoordinator is from dispatces.workflow.coordinator, not from idaes.apps.grid_integration.
# This version contains some modifications for generator typing and will be merged into idaes in the future.
###

this_file_path = Path(this_file_dir())

usage = "Run double loop simulation with RE model."
parser = ArgumentParser(usage)

parser.add_argument(
    "--sim_id",
    dest="sim_id",
    help="Indicate the simulation ID.",
    action="store",
    type=int,
    default=0,
)

parser.add_argument(
    "--wind_pmax",
    dest="wind_pmax",
    help="Set wind capacity in MW.",
    action="store",
    type=float,
    default=fixed_wind_mw,
)

parser.add_argument(
    "--pem_pmax",
    dest="pem_pmax",
    help="Set the PEM power capacity in MW.",
    action="store",
    type=float,
    default=127.05,
)

parser.add_argument(
    "--pem_bid",
    dest="pem_bid",
    help="Set the PEM bid price in $/MW.",
    action="store",
    type=float,
    default=15.0,
)

parser.add_argument(
    "--reserve_factor",
    dest="reserve_factor",
    help="Set the reserve factor.",
    action="store",
    type=float,
    default=0.15,
)

parser.add_argument(
    "--shortfall",
    dest="shortfall",
    help="Set the shortfall price.",
    action="store",
    type=float,
    default=1000,
)

options = parser.parse_args()

sim_id = options.sim_id
wind_pmax = options.wind_pmax
pem_pmax = options.pem_pmax
pem_bid = options.pem_bid
reserve_factor = options.reserve_factor
shortfall = options.shortfall

p_min = 0
default_wind_bus = 303
bus_name = "Caesar"
wind_generator = "303_WIND_1"
start_date = "01-01-2020"
n_days = 366

# NOTE: `rts_gmlc_data_dir` should point to a directory containing RTS-GMLC scenarios
rts_gmlc_data_dir = rts_gmlc.source_data_path
wind_df = read_rts_gmlc_wind_inputs(rts_gmlc_data_dir, wind_generator)
wind_df = wind_df[wind_df.index >= start_date]
wind_rt_cfs = wind_df[f"{wind_generator}-RTCF"].values.tolist()

output_dir = Path(f"double_loop_parametrized_results")

solver = pyo.SolverFactory("xpress_direct")

thermal_generator_params = {
    "gen_name": wind_generator,
    "bus": bus_name,
    "p_min": p_min,
    "p_max": wind_pmax,
    "min_down_time": 0,
    "min_up_time": 0,
    "ramp_up_60min": wind_pmax,
    "ramp_down_60min": wind_pmax,
    "shutdown_capacity": wind_pmax,
    "startup_capacity": wind_pmax,
    "initial_status": 1,                                        # Has been off for 1 hour before start of simulation
    "initial_p_output": 0,
    "production_cost_bid_pairs": [(p_min, 0), (wind_pmax, 0)],
    "include_default_p_cost": False,
    "startup_cost_pairs": [(0, 0)],
    "fixed_commitment": 1,                                      # Same as the plant in the parameter sweep, which was RE-type and always on
    "spinning_capacity": 0,                                     # Disable participation in some reserve services
    "non_spinning_capacity": 0,
    "supplemental_spinning_capacity": 0,
    "supplemental_non_spinning_capacity": 0,
    "agc_capable": False
}
model_data = ThermalGeneratorModelData(**thermal_generator_params)


################################################################################
################################# bidder #######################################
################################################################################

# same horizons as parameter sweep
day_ahead_horizon = 36                                          
real_time_horizon = 1

# PerfectForecaster uses Dataframe with RT and DA wind resource and LMPs. However, for the ParametrizedBidder that
# makes the bid curve based on the `pem_bid` parameter, the LMPs are not needed and are never accessed.
# For non-RE plants, the CFs are never accessed.
forecaster = PerfectForecaster(wind_df)

mp_wind_pem_bid = MultiPeriodWindPEM(
    model_data=model_data,
    wind_capacity_factors=wind_rt_cfs,
    wind_pmax_mw=wind_pmax,
    pem_pmax_mw=pem_pmax
)

# ParametrizedBidder is an alternative to the stochastic LMP problem, and serves up a bid curve per timestep
# that is a function of only the bid parameter, PEM capacity and wind resource (DA or RT)
bidder_object = PEMParametrizedBidder(
    bidding_model_object=mp_wind_pem_bid,
    day_ahead_horizon=day_ahead_horizon,
    real_time_horizon=real_time_horizon,
    solver=solver,
    forecaster=forecaster,
    pem_marginal_cost=pem_bid,
    pem_mw=pem_pmax
)

################################################################################
################################# Tracker ######################################
################################################################################

# same tracking_horizon as parameter sweep and n_tracking_hour does not need to be >1
tracking_horizon = 4
n_tracking_hour = 1

mp_wind_pem_track = MultiPeriodWindPEM(
    model_data=model_data,
    wind_capacity_factors=wind_rt_cfs,
    wind_pmax_mw=wind_pmax,
    pem_pmax_mw=pem_pmax
)

# create a `Tracker` using`mp_wind_battery`
tracker_object = Tracker(
    tracking_model_object=mp_wind_pem_track,
    tracking_horizon=tracking_horizon,
    n_tracking_hour=n_tracking_hour,
    solver=solver,
)

mp_wind_pem_track_project = MultiPeriodWindPEM(
    model_data=model_data,
    wind_capacity_factors=wind_rt_cfs,
    wind_pmax_mw=wind_pmax,
    pem_pmax_mw=pem_pmax
)

# create a `Tracker` using`mp_wind_battery`
project_tracker_object = Tracker(
    tracking_model_object=mp_wind_pem_track_project,
    tracking_horizon=tracking_horizon,
    n_tracking_hour=n_tracking_hour,
    solver=solver,
)

################################################################################
################################# Coordinator ##################################
################################################################################

coordinator = DoubleLoopCoordinator(
    bidder=bidder_object,
    tracker=tracker_object,
    projection_tracker=project_tracker_object,
)

# same options as for prescient sweep
prescient_options = {
    "data_path": rts_gmlc_data_dir,
    "reserve_factor": reserve_factor,
    "simulate_out_of_sample": True,
    "output_directory": output_dir,
    "monitor_all_contingencies":False,
    "input_format": "rts-gmlc",
    "start_date": start_date,
    "num_days": n_days,
    "sced_horizon": real_time_horizon,
    "ruc_mipgap": 0.01,
    "deterministic_ruc_solver": "xpress_persistent",
    "deterministic_ruc_solver_options" : {"threads":2, "heurstrategy":2, "cutstrategy":3, "symmetry":2, "maxnode":1000},
    "sced_solver": "xpress_persistent",
    "sced_frequency_minutes":60,
	    "sced_solver_options" : {"threads":1},
    "ruc_horizon": day_ahead_horizon,
    "compute_market_settlements": True,
    "price_threshold": shortfall,
    "transmission_price_threshold": shortfall / 2,
    "contingency_price_threshold":None,
    "reserve_price_threshold": shortfall / 10,
    "day_ahead_pricing": "aCHP",
    "enforce_sced_shutdown_ramprate":False,
    "ruc_slack_type":"ref-bus-and-branches",    # slack var power balance at reference bus and transmission line flows vs slack var for power balance at every bus
    "sced_slack_type":"ref-bus-and-branches",
    "disable_stackgraphs":True,
    "symbolic_solver_labels": True,
    "plugin": {
        "doubleloop": {
            "module": coordinator.prescient_plugin_module,
            "bidding_generator": wind_generator,
        }
    },
    # verbosity, turn on for debugging
    "output_ruc_solutions": False,
    "write_deterministic_ruc_instances": False,
    "write_sced_instances": False,
    "print_sced": False
}

Prescient().simulate(**prescient_options)

# write options into the result folder
with open(output_dir / "sim_options.json", "w") as f:
    f.write(str(thermal_generator_params.update(prescient_options)))


#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
from idaes.apps.grid_integration.bidder import *
from dispatches.workflow.parametrized_bidder import PerfectForecaster, ParametrizedBidder


class PEMParametrizedBidder(ParametrizedBidder):
    """
    Wind + PEM bidder that uses parameterized bid curve.

    'pem_marginal_cost': the cost/MW above which all available wind energy will be sold to grid;
        below which, make hydrogen and sell remainder of wind to grid
    'pem_mw': maximum PEM capacity limits how much energy is bid at the `pem_marginal_cost`

    Every timestep for RT or DA, max energy bid is the available wind resource
    """

    def __init__(
        self,
        bidding_model_object,
        day_ahead_horizon,
        real_time_horizon,
        solver,
        forecaster,
        pem_marginal_cost,
        pem_mw
    ):
        super().__init__(bidding_model_object,
                         day_ahead_horizon,
                         real_time_horizon,
                         solver,
                         forecaster)
        self.wind_marginal_cost = 0
        self.wind_mw = self.bidding_model_object._wind_pmax_mw
        self.pem_marginal_cost = pem_marginal_cost
        self.pem_mw = pem_mw

    def compute_day_ahead_bids(self, date, hour=0):
        """
        DA Bid: from 0 MW to (Wind Resource - PEM capacity) MW, bid $0/MWh.
        from (Wind Resource - PEM capacity) MW to Wind Resource MW, bid 'pem_marginal_cost'

        If Wind resource at some time is less than PEM capacity, then reduce to available resource
        """
        gen = self.generator
        forecast = self.forecaster.forecast_day_ahead_capacity_factor(date, hour, gen, self.day_ahead_horizon)

        full_bids = {}

        for t_idx in range(self.day_ahead_horizon):
            da_wind = forecast[t_idx] * self.wind_mw
            grid_wind = max(0, da_wind - self.pem_mw)
            bids = [(0, 0), (grid_wind, 0), (da_wind, self.pem_marginal_cost)]
            cost_curve = convert_marginal_costs_to_actual_costs(bids)

            temp_curve = {
                    "data_type": "cost_curve",
                    "cost_curve_type": "piecewise",
                    "values": cost_curve,
            }
            tx_utils.validate_and_clean_cost_curve(
                curve=temp_curve,
                curve_type="cost_curve",
                p_min=0,
                p_max=max([p[0] for p in cost_curve]),
                gen_name=gen,
                t=t_idx,
            )

            t = t_idx + hour
            full_bids[t] = {}
            full_bids[t][gen] = {}
            full_bids[t][gen]["p_cost"] = cost_curve
            full_bids[t][gen]["p_min"] = 0
            full_bids[t][gen]["p_max"] = da_wind
            full_bids[t][gen]["startup_capacity"] = da_wind
            full_bids[t][gen]["shutdown_capacity"] = da_wind

        self._record_bids(full_bids, date, hour, Market="Day-ahead")
        return full_bids

    def compute_real_time_bids(
        self, date, hour, _, __
    ):
        """
        RT Bid: from 0 MW to (Wind Resource - PEM capacity) MW, bid $0/MWh.
        from (Wind Resource - PEM capacity) MW to Wind Resource MW, bid 'pem_marginal_cost'

        If Wind resource at some time is less than PEM capacity, then reduce to available resource
        """
        gen = self.generator
        forecast = self.forecaster.forecast_real_time_capacity_factor(date, hour, gen, self.day_ahead_horizon)
        
        full_bids = {}

        for t_idx in range(self.real_time_horizon):
            rt_wind = forecast[t_idx] * self.wind_mw
            grid_wind = max(0, rt_wind - self.pem_mw)
            bids = [(0, 0), (grid_wind, 0), (rt_wind, self.pem_marginal_cost)]

            t = t_idx + hour
            full_bids[t] = {}
            full_bids[t][gen] = {}
            full_bids[t][gen]["p_cost"] = convert_marginal_costs_to_actual_costs(bids)
            full_bids[t][gen]["p_min"] = 0
            full_bids[t][gen]["p_max"] = rt_wind
            full_bids[t][gen]["startup_capacity"] = rt_wind
            full_bids[t][gen]["shutdown_capacity"] = rt_wind

        self._record_bids(full_bids, date, hour, Market="Real-time")
        return full_bids



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
import numpy as np
from idaes.apps.grid_integration.bidder import *
from idaes.apps.grid_integration.forecaster import AbstractPrescientPriceForecaster


class PerfectForecaster(AbstractPrescientPriceForecaster):
    def __init__(self, data_path_or_df):
        """
        Perfect forecaster that reads the data from a Dataframe containing:
         - {bus}-DALMP
         - {bus}-RTLMP
         - {bus}-DACF and {bus}-RTCF for renewable plants
        """
        if isinstance(data_path_or_df, str):
            self.data = pd.read_csv(data_path_or_df, index_col="Datetime", parse_dates=True)
        elif isinstance(data_path_or_df, pd.DataFrame):
            self.data = data_path_or_df
        else:
            raise ValueError

    def __getitem__(self, index):
        return self.data[index]

    def fetch_hourly_stats_from_prescient(self, prescient_hourly_stats):
        pass

    def fetch_day_ahead_stats_from_prescient(self, uc_date, uc_hour, day_ahead_result):
        pass

    def forecast_day_ahead_and_real_time_prices(self, date, hour, bus, horizon, _):
        rt_forecast = self.forecast_real_time_prices(
            date, hour, bus, horizon, _
        )
        da_forecast = self.forecast_day_ahead_prices(
            date, hour, bus, horizon, _
        )
        return da_forecast, rt_forecast

    def get_column_from_data(self, date, hour, horizon, col):
        datetime_index = pd.to_datetime(date) + pd.Timedelta(hours=hour)
        forecast = self.data[self.data.index >= datetime_index].head(horizon)
        values = forecast[col].values
        if len(values) < horizon:
            values = np.append(values, self.data[col].values[:horizon - len(values)])
        return values

    def forecast_day_ahead_prices(self, date, hour, bus, horizon, _):
        return self.get_column_from_data(date, hour, horizon, f'{bus}-DALMP')

    def forecast_real_time_prices(self, date, hour, bus, horizon, _):
        return self.get_column_from_data(date, hour, horizon, f'{bus}-RTLMP')

    def forecast_day_ahead_capacity_factor(self, date, hour, gen, horizon):
        return self.get_column_from_data(date, hour, horizon, f'{gen}-DACF')

    def forecast_real_time_capacity_factor(self, date, hour, gen, horizon):
        return self.get_column_from_data(date, hour, horizon, f'{gen}-RTCF')


class ParametrizedBidder(AbstractBidder):

    """
    Template class for bidders that uses parameters to create the DA and RT bid curves
    """
    def __init__(
        self,
        bidding_model_object,
        day_ahead_horizon,
        real_time_horizon,
        solver,
        forecaster,
    ):
        self.bidding_model_object = bidding_model_object
        self.day_ahead_horizon = day_ahead_horizon
        self.real_time_horizon = real_time_horizon
        self.n_scenario = 1
        self.solver = solver
        self.forecaster = forecaster
        self.real_time_underbid_penalty = 500

        self._check_inputs()

        self.generator = self.bidding_model_object.model_data.gen_name

        self.bids_result_list = []

        self.battery_marginal_cost = 25
        self.battery_capacity_ratio = 0.4

    def formulate_DA_bidding_problem(self):
        pass

    def formulate_RT_bidding_problem(self):
        pass

    def compute_day_ahead_bids(self, date, hour=0):
        raise NotImplementedError

    def compute_real_time_bids(
        self, date, hour, realized_day_ahead_prices, realized_day_ahead_dispatches, tracker_profile
    ):
        raise NotImplementedError

    def update_real_time_model(self, **kwargs):
        pass

    def update_day_ahead_model(self, **kwargs):
        pass

    def record_bids(self, bids, model, date, hour, market):

        """
        This function records the bids and the details in the underlying bidding model.

        Arguments:
            bids: the obtained bids for this date.

            model: bidding model

            date: the date we bid into

            hour: the hour we bid into

        Returns:
            None

        """

        # record bids
        self._record_bids(bids, date, hour, Market=market)

        # record the details of bidding model
        for i in model.SCENARIOS:
            self.bidding_model_object.record_results(
                model.fs[i], date=date, hour=hour, Scenario=i, Market=market
            )

        return

    def _record_bids(self, bids, date, hour, **kwargs):
        df_list = []
        for t in bids:
            for gen in bids[t]:

                result_dict = {}
                result_dict["Generator"] = gen
                result_dict["Date"] = date
                result_dict["Hour"] = t

                for k, v in kwargs.items():
                    result_dict[k] = v

                pair_cnt = len(bids[t][gen]["p_cost"])

                for idx, (power, cost) in enumerate(bids[t][gen]["p_cost"]):
                    result_dict[f"Power {idx} [MW]"] = power
                    result_dict[f"Cost {idx} [$]"] = cost

                # place holder, in case different len of bids
                while pair_cnt < self.n_scenario:
                    result_dict[f"Power {pair_cnt} [MW]"] = None
                    result_dict[f"Cost {pair_cnt} [$]"] = None

                    pair_cnt += 1

                result_df = pd.DataFrame.from_dict(result_dict, orient="index")
                df_list.append(result_df.T)

        # save the result to object property
        # wait to be written when simulation ends
        self.bids_result_list.append(pd.concat(df_list))

        return

    def write_results(self, path):
        """
        This methods writes the saved operation stats into an csv file.

        Arguments:
            path: the path to write the results.

        Return:
            None
        """

        print("")
        print("Saving bidding results to disk...")
        pd.concat(self.bids_result_list).to_csv(
            os.path.join(path, "bidder_detail.csv"), index=False
        )
        return

    @property
    def generator(self):
        return self._generator

    @generator.setter
    def generator(self, name):
        self._generator = name




#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
from pathlib import Path
import pickle

import pytest

keras = pytest.importorskip("keras", reason=f"keras required to run {__file__}")

from dispatches.case_studies.renewables_case.RE_surrogate_optimization_steadystate import re_nn_dir


@pytest.fixture(scope="module")
def base_dir() -> Path:
    return re_nn_dir


def test_base_dir(base_dir: Path):
    assert base_dir.is_dir()
    contents = sorted(base_dir.rglob("*"))
    assert len(contents) > 0


class TestModelDeserialization:

    @pytest.mark.parametrize(
        "name",
        [
            "RE_revenue_2_25",
        ]
    )
    def test_revenue(self, base_dir: Path, name: str):
        model = keras.models.load_model(base_dir / "revenue" / name)

        from keras.engine.sequential import Sequential
        assert isinstance(model, Sequential)

    @pytest.mark.parametrize(
        "name",
        [
            "ss_surrogate_model_wind_pmax",
        ]
    )
    def test_dispatch(self, base_dir: Path, name: str):
        model = keras.models.load_model(base_dir / "dispatch_frequency" / name)

        from keras.engine.sequential import Sequential
        assert isinstance(model, Sequential)

    def test_clustering(self, base_dir: Path):
        with (base_dir / "dispatch_frequency" / "static_clustering_wind_pmax.pkl").open("rb") as f:
            model = pickle.load(f)

        from sklearn.cluster import KMeans
        assert isinstance(model, KMeans)


#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
import pytest
from pathlib import Path
import pandas as pd
import os
import numpy as np
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.apps.grid_integration.tracker import Tracker
from idaes.apps.grid_integration.bidder import SelfScheduler, Bidder
from idaes.apps.grid_integration.forecaster import Backcaster
from idaes.apps.grid_integration.model_data import (
    RenewableGeneratorModelData,
    ThermalGeneratorModelData,
)
from dispatches.case_studies.renewables_case.wind_PEM_double_loop import MultiPeriodWindPEM, update_wind_capacity_factor
from dispatches.case_studies.renewables_case.PEM_parametrized_bidder import PEMParametrizedBidder, PerfectForecaster


re_case_dir = Path(this_file_dir()).parent
wind_thermal_dispatch_data = pd.read_csv(re_case_dir / "data" / "Wind_Thermal_Dispatch.csv")
wind_thermal_dispatch_data.index = pd.to_datetime(wind_thermal_dispatch_data["DateTime"])

tracking_horizon = 4
day_ahead_horizon = 48
real_time_horizon = 4
pmin = 0
pmax = 200
bus_name = "Carter"

generator_params = {
    "gen_name": "309_WIND_1",
    "bus": bus_name,
    "p_min": pmin,
    "p_max": pmax,
    "p_cost": 0,
    "fixed_commitment": None,
}


def test_track_market_dispatch():
    solver = pyo.SolverFactory("ipopt")

    model_data = RenewableGeneratorModelData(**generator_params)

    mp_wind_battery = MultiPeriodWindPEM(
        model_data=model_data,
        wind_capacity_factors=wind_thermal_dispatch_data["309_WIND_1-RTCF"].values,
        wind_pmax_mw=pmax,
        pem_pmax_mw=25
    )

    n_tracking_hour = 1

    # create a `Tracker` using`mp_wind_battery`
    tracker_object = Tracker(
        tracking_model_object=mp_wind_battery,
        tracking_horizon=tracking_horizon,
        n_tracking_hour=n_tracking_hour,
        solver=solver,
    )

    cap_factors = mp_wind_battery._get_capacity_factors(tracker_object.model.fs)
    assert cap_factors[0] == pytest.approx(0.00562, rel=1e-3)

    # example market dispatch signal for 4 hours
    market_dispatch = [0, 1.5, 15.0, 24.5]

    # find a solution that tracks the dispatch signal
    tracker_object.track_market_dispatch(
        market_dispatch=market_dispatch, date="2020-01-02", hour="00:00"
    )

    blks = tracker_object.model.fs.windPEM.get_active_process_blocks()
    assert len(blks) == tracking_horizon

    wind_power = [blks[i].fs.windpower.electricity[0].value for i in range(4)]
    wind_waste = [pyo.value(tracker_object.model.fs.wind_waste[i]) for i in range(4)]
    expected_wind_power = [1123.85, 1573.38, 20510, 25938]

    for power, expected in zip(wind_power, expected_wind_power):
        assert (
            pytest.approx(power, rel=1e-3) == expected
        )
    for power in wind_waste:
        assert (
            pytest.approx(power, abs=1e-3) == 0
        )

    produced_power = [pyo.value(tracker_object.power_output[t])
                      for t in range(tracking_horizon)]

    for power, expected in zip(produced_power, market_dispatch):
        assert (
            pytest.approx(power, abs=1e-3) == expected
        )

    expected_pem_energy = [expected_wind_power[i] -
                               market_dispatch[i] * 1e3 for i in range(tracking_horizon)]
    pem_power = [blks[i].fs.pem.electricity[0].value for i in range(4)]

    for power, expected in zip(pem_power, expected_pem_energy):
        assert (
            pytest.approx(power, rel=1e-3) == expected
        )

    mp_wind_battery.update_model(tracker_object.model.fs, [0] * 4)


def test_compute_parametrized_bids_DA():
    forecaster = PerfectForecaster(wind_thermal_dispatch_data)

    model_data = RenewableGeneratorModelData(**generator_params)

    solver = pyo.SolverFactory("cbc")

    mp_wind_battery_bid = MultiPeriodWindPEM(
        model_data=model_data,
        wind_capacity_factors=wind_thermal_dispatch_data["309_WIND_1-RTCF"].values,
        wind_pmax_mw=pmax,
        pem_pmax_mw=25
    )

    bidder_object = PEMParametrizedBidder(
        bidding_model_object=mp_wind_battery_bid,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        solver=solver,
        forecaster=forecaster,
        pem_marginal_cost=30,
        pem_mw=25
    )

    date = "2020-01-02"
    bid_energies = bidder_object.compute_day_ahead_bids(date=date)
    bid_energies = [i['309_WIND_1']['p_max'] for i in bid_energies.values()]

    known_solution = [0.13, 1.08, 3.64, 15.37, 24.68, 31.83, 33.18, 13.89, 7.55, 4.99, 1.08, 
                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.81, 1.89, 33.45, 34.12, 17.8, 13.49, 
                      48.15, 69.72, 36.01, 111.67, 132.7, 178.56, 107.89, 146.86, 145.38, 143.9, 
                      156.84, 111.94, 87.66, 154.69, 79.3, 116.12, 111.8, 98.18, 98.58, 106.54, 
                      130.14, 124.75, 160.08, 153.47, 162.1]

    assertStructuredAlmostEqual(bid_energies, known_solution, abstol=1e-2)


def test_compute_parametrized_bids_RT():
    wind_pmax = 200
    pem_pmax = 25

    forecaster = PerfectForecaster(wind_thermal_dispatch_data)

    thermal_generator_params = {
        "gen_name": "309_WIND_1",
        "bus": bus_name,
        "p_min": pmin,
        "p_max": wind_pmax,
        "min_down_time": 0,
        "min_up_time": 0,
        "ramp_up_60min": wind_pmax + pem_pmax,
        "ramp_down_60min": wind_pmax + pem_pmax,
        "shutdown_capacity": wind_pmax + pem_pmax,
        "startup_capacity": 0,
        "initial_status": 1,
        "initial_p_output": 0,
        "production_cost_bid_pairs": [(pmin, 0), (wind_pmax, 0)],
        "include_default_p_cost": False,
        "startup_cost_pairs": [(0, 0)],
        "fixed_commitment": None,
    }
    model_data = ThermalGeneratorModelData(**thermal_generator_params)

    solver = pyo.SolverFactory("cbc")

    mp_wind_battery_bid = MultiPeriodWindPEM(
        model_data=model_data,
        wind_capacity_factors=wind_thermal_dispatch_data["309_WIND_1-RTCF"].values,
        wind_pmax_mw=wind_pmax,
        pem_pmax_mw=pem_pmax
    )

    bidder_object = PEMParametrizedBidder(
        bidding_model_object=mp_wind_battery_bid,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        solver=solver,
        forecaster=forecaster,
        pem_marginal_cost=30,
        pem_mw=25
    )

    date = "2020-01-02"
    bids = bidder_object.compute_real_time_bids(date=date, hour=0, _=None, __=None)
    bids = [i['309_WIND_1']['p_cost'] for i in bids.values()]
    bid_prices = [bid[-1][1] for bid in bids]

    known_solution = [33.72, 47.2, 615.31, 750.0]

    assertStructuredAlmostEqual(bid_prices, known_solution, reltol=1e-2)

    bidder_object.write_results(this_file_dir())
    os.remove(Path(this_file_dir()) / "bidder_detail.csv")